from copy import deepcopy

import numpy as np
from rdkit import Chem
from rdkit import DataStructs
from rdkit.Chem.AtomPairs import Pairs
from rdkit.Chem.Scaffolds import MurckoScaffold

from reinvent_scoring.scoring.diversity_filters.reinvent_core.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.reinvent_core.diversity_filter_parameters import DiversityFilterParameters
from reinvent_scoring.scoring.score_summary import FinalSummary


class ScaffoldSimilarity(BaseDiversityFilter):
    """Penalizes compounds based on atom pair Tanimoto similarity to previously generated Murcko Scaffolds."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)
        self._scaffold_fingerprints = {}

    def update_score(self, score_summary: FinalSummary, step=0) -> np.array:
        score_summary = deepcopy(score_summary)
        scores = score_summary.total_score
        smiles = score_summary.scored_smiles

        for i in score_summary.valid_idxs:
            smile = self._chemistry.convert_to_rdkit_smiles(smiles[i])
            scaffold = self._calculate_scaffold(smile)

            # check, if another scaffold should be used as "bucket", because it is very similar as defined by the
            # "minsimilarity" threshold; if not, this call is a no-op and the smiles' normal Murcko scaffold will be used in case
            # -> usage of the "murcko scaffold filter" is actually a special case, where "minsimilarity" is 1.0
            scaffold = self._find_similar_scaffold(scaffold)

            scores[i] = 0 if self._smiles_exists(smile) else scores[i]
            if scores[i] >= self.parameters.minscore:
                self._add_to_memory(i, scores[i], smile, scaffold, score_summary.scaffold_log, step)
                scores[i] = self._penalize_score(scaffold, scores[i])
        return scores

    def _calculate_scaffold(self, smile):
        mol = Chem.MolFromSmiles(smile)
        if mol:
            try:
                scaffold = MurckoScaffold.GetScaffoldForMol(mol)
                return Chem.MolToSmiles(scaffold, isomericSmiles=False)
            except ValueError:
                scaffold_smiles = ''
        else:
            scaffold_smiles = ''
        return scaffold_smiles

    def _find_similar_scaffold(self, scaffold):
        """
        this function tries to find a "similar" scaffold (according to the threshold set by parameter "minsimilarity") and if at least one
        scaffold satisfies this criteria, it will replace the smiles' scaffold with the most similar one
        -> in effect, this reduces the number of scaffold buckets in the memory (the lower parameter "minsimilarity", the more
           pronounced the reduction)
        generate a "mol" scaffold from the smile and calculate an atom pair fingerprint

        :param scaffold: scaffold represented by a smiles string
        :return: closest scaffold given a certain similarity threshold 
        """
        if scaffold is not '':
            fp = Pairs.GetAtomPairFingerprint(Chem.MolFromSmiles(scaffold))

            # make a list of the stored fingerprints for similarity calculations
            fps = list(self._scaffold_fingerprints.values())

            # check, if a similar scaffold entry already exists and if so, use this one instead
            if len(fps) > 0:
                similarity_scores = DataStructs.BulkDiceSimilarity(fp, fps)
                closest = np.argmax(similarity_scores)
                if similarity_scores[closest] >= self.parameters.minsimilarity:
                    scaffold = list(self._scaffold_fingerprints.keys())[closest]
                    fp = self._scaffold_fingerprints[scaffold]

            self._scaffold_fingerprints[scaffold] = fp
        return scaffold




import pickle

from typing import List

from reinvent_scoring.scoring.predictive_model.model_container import ModelContainer
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary
from reinvent_scoring.scoring.score_transformations import TransformationFactory
from reinvent_scoring.scoring.enums import TransformationTypeEnum


class PredictivePropertyComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self.activity_model = self._load_model(parameters)
        self._transformation_function = self._assign_transformation(parameters.specific_parameters)

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        score, raw_score = self._predict_and_transform(molecules)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)
        return score_summary

    def _predict_and_transform(self, molecules: List):
        score = self.activity_model.predict(molecules, self.parameters.specific_parameters)
        transformed_score = self._apply_transformation(score, self.parameters.specific_parameters)
        return transformed_score, score

    def _load_model(self, parameters: ComponentParameters):
        try:
            activity_model = self._load_container(parameters)
        except:
            model_path = self.parameters.specific_parameters.get(self.component_specific_parameters.MODEL_PATH, "")
            raise Exception(f"The loaded file `{model_path}` isn't a valid scikit-learn model")
        return activity_model

    def _load_container(self, parameters: ComponentParameters):
        model_path = self.parameters.specific_parameters.get(self.component_specific_parameters.MODEL_PATH, "")
        with open(model_path, "rb") as f:
            scikit_model = pickle.load(f)
            packaged_model = ModelContainer(scikit_model, parameters.specific_parameters)
        return packaged_model

    def _apply_transformation(self, predicted_activity, parameters: dict):
        if parameters.get(self.component_specific_parameters.TRANSFORMATION, False):
            activity = self._transformation_function(predicted_activity, parameters)
        else:
            activity = predicted_activity
        return activity

    def _assign_transformation(self, specific_parameters: dict):
        transformation_type = TransformationTypeEnum()
        """classification models should not have any prediction transformations"""
        if specific_parameters[self.component_specific_parameters.SCIKIT] == "classification":
            specific_parameters[self.component_specific_parameters.TRANSFORMATION] = False
            specific_parameters[self.component_specific_parameters.TRANSFORMATION_TYPE] = transformation_type.NO_TRANSFORMATION
        factory = TransformationFactory()
        transform_function = factory.get_transformation_function(specific_parameters)
        return transform_function



from collections import namedtuple

from openeye import oeshape

SIM_FUNC = namedtuple('SIM_FUNC', ['shape', 'color', 'predicate'])

def get_similarity_name_set(parameters, param_names_enum, sim_measure_enum):
    similarity_collection_name = parameters.specific_parameters.get(param_names_enum.SIM_MEASURE,
                                                                         sim_measure_enum.TANIMOTO)
    return similarity_collection(similarity_collection_name, sim_measure_enum)


def similarity_collection(sim_measure_type, sim_measure_enum):
    sim_def_dict = {
        sim_measure_enum.TANIMOTO: SIM_FUNC('GetTanimoto', 'GetColorTanimoto', 'OEHighestTanimotoCombo'),
        sim_measure_enum.REF_TVERSKY: SIM_FUNC('GetRefTversky', 'GetRefColorTversky',
                                                     'OEHighestRefTverskyCombo'),
        sim_measure_enum.FIT_TVERSKY: SIM_FUNC('GetFitTversky', 'GetFitColorTversky',
                                                     'OEHighestFitTverskyCombo'),
    }
    return sim_def_dict.get(sim_measure_type)

def init_cff(prep, cff_path):
    if len(cff_path) == 0:
        cff_path = oeshape.OEColorFFType_ImplicitMillsDean
    cff = oeshape.OEColorForceField()
    if cff.Init(cff_path):
        prep.SetColorForceField(cff)
    else:
        raise ValueError("Custom color force field initialisation failed")


from abc import ABC, abstractmethod

from typing import Dict, List


class BaseModelContainer(ABC):

    @abstractmethod
    def predict(self, molecules: List, parameters: Dict):
        raise NotImplementedError("'predict' method is not implemented !")



from typing import List

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import TanimotoSimilarity, \
    JaccardDistance, CustomAlerts, QedScore, MatchingSubstructure, \
    RocsSimilarity, ParallelRocsSimilarity, PredictivePropertyComponent, SelectivityComponent, \
    SASComponent, MolWeight, PSA, RotatableBonds, HBD_Lipinski, HBA_Lipinski, \
    NumRings, AZlogD74, HLMClint, SlogP, \
    RHClint, HHClint, SolubilityDD, HERG, CACO2Intrinsic, CACO2Efflux, AZdock, RatPKPiP, Top20, GraphLength, \
    NumberOfStereoCenters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.enums import ScoringFunctionComponentNameEnum
from reinvent_scoring.scoring.score_components.pip.pip_log_prediction_component import PiPLogPredictionComponent
from reinvent_scoring.scoring.score_components.pip.pip_prediction_component import PiPPredictionComponent


class ScoreComponentFactory:
    def __init__(self, parameters: List[ComponentParameters]):
        self._parameters = parameters
        self._current_components = self._deafult_scoring_component_registry()

    def _deafult_scoring_component_registry(self) -> dict:
        enum = ScoringFunctionComponentNameEnum()
        component_map = {
            enum.MATCHING_SUBSTRUCTURE: MatchingSubstructure,
            enum.ROCS_SIMILARITY: RocsSimilarity,
            enum.PREDICTIVE_PROPERTY: PredictivePropertyComponent,
            enum.TANIMOTO_SIMILARITY: TanimotoSimilarity,
            enum.JACCARD_DISTANCE: JaccardDistance,
            enum.CUSTOM_ALERTS: CustomAlerts,
            enum.QED_SCORE: QedScore,
            enum.MOLECULAR_WEIGHT: MolWeight,
            enum.TPSA: PSA,
            enum.NUM_ROTATABLE_BONDS: RotatableBonds,
            enum.GRAPH_LENGTH: GraphLength,
            enum.NUM_HBD_LIPINSKI: HBD_Lipinski,
            enum.NUM_HBA_LIPINSKI: HBA_Lipinski,
            enum.NUM_RINGS: NumRings,
            enum.SLOGP: SlogP,
            enum.NUMBER_OF_STEREO_CENTERS: NumberOfStereoCenters,
            enum.PARALLEL_ROCS_SIMILARITY: ParallelRocsSimilarity,
            enum.AZ_LOGD74: AZlogD74,
            enum.HLM_CLINT: HLMClint,
            enum.RH_CLINT: RHClint,
            enum.HH_CLINT: HHClint,
            enum.SOLUBILITY_DD: SolubilityDD,
            enum.HERG: HERG,
            enum.CACO2_INTR: CACO2Intrinsic,
            enum.CACO2_EFFLUX: CACO2Efflux,
            enum.SELECTIVITY: SelectivityComponent,
            enum.SA_SCORE: SASComponent,
            enum.AZDOCK: AZdock,
            enum.AZ_LOGD74_PIP: PiPPredictionComponent,
            enum.CACO2_INTR_PIP: PiPLogPredictionComponent,
            enum.CACO2_EFFLUX_PIP: PiPPredictionComponent,
            enum.HH_CLINT_PIP: PiPLogPredictionComponent,
            enum.HLM_CLINT_PIP: PiPLogPredictionComponent,
            enum.RH_CLINT_PIP: PiPLogPredictionComponent,
            enum.SOLUBILITY_DD_PIP: PiPLogPredictionComponent,
            enum.HERG_PIP: PiPPredictionComponent,
            enum.RAT_PK_PIP: RatPKPiP,
            enum.CLAB_TOP_20: Top20,
            enum.RA_SCORE: PiPPredictionComponent,
            enum.KPUU_PIP: PiPLogPredictionComponent
        }
        return component_map

    def create_score_components(self) -> [BaseScoreComponent]:
        return [self._current_components.get(p.component_type)(p) for p in self._parameters]



from dataclasses import dataclass

import numpy as np
from typing import List

from reinvent_scoring.scoring.component_parameters import ComponentParameters


@dataclass
class ComponentSummary:
    total_score: np.array
    parameters: ComponentParameters
    raw_score: np.ndarray = None


class FinalSummary:
    def __init__(self, total_score: np.array, scored_smiles: List[str], valid_idxs: List[int],
                 scaffold_log_summary: List[ComponentSummary]):
        self.total_score = total_score
        self.scored_smiles = scored_smiles
        self.valid_idxs = valid_idxs
        score = [LoggableComponent(c.parameters.component_type, c.parameters.name, c.total_score) for c in scaffold_log_summary]
        raw_score = [LoggableComponent(c.parameters.component_type, f'raw_{c.parameters.name}', c.raw_score) for c in
                     scaffold_log_summary if c.raw_score is not None]
        score.extend(raw_score)
        self.scaffold_log: List[ComponentSummary] = scaffold_log_summary
        self.profile: List[LoggableComponent] = score




@dataclass
class LoggableComponent:
    component_type: str
    name: str
    score: np.array



import io
import subprocess
from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class BaseStructuralComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def calculate_score_for_step(self, molecules: List, step=-1) -> ComponentSummary:
        return self.calculate_score(molecules, step)

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score, raw_score = self._calculate_score(valid_smiles, step)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)
        return score_summary

    def _get_step_string(self, step) -> str:
        if step == -1:
            return "\"\""
        return "".join(["\"e", str(step).zfill(4), "_\""])

    @abstractmethod
    def _calculate_score(self, smiles: List[str], step) -> np.array:
        raise NotImplementedError("_calculate_score method is not implemented")

    @abstractmethod
    def _create_command(self, input_file, step) -> str:
        raise NotImplementedError("_create_command method is not implemented")

    def _send_request_with_stepwize_read(self, command, data_size: int):
        with subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              shell=True) as proc:
            wrapped_proc_in = io.TextIOWrapper(proc.stdin, 'utf-8')
            wrapped_proc_out = io.TextIOWrapper(proc.stdout, 'utf-8')
            result = [self._parse_result(wrapped_proc_out.readline()) for i in range(data_size)]
            wrapped_proc_in.close()
            wrapped_proc_out.close()
            proc.wait()
            proc.terminate()
        return result



from typing import List

from reinvent_chemistry.similarity import Similarity
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class JaccardDistance(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._similarity = Similarity()
        self._radius = self.parameters.specific_parameters.get("radius", 3)
        self._use_counts = self.parameters.specific_parameters.get("use_counts", True)
        self._use_features = self.parameters.specific_parameters.get("use_features", True)
        smiles = self.parameters.specific_parameters.get(self.component_specific_parameters.SMILES, [])
        self._ref_fingerprints = self._chemistry.smiles_to_fingerprints(smiles, radius=self._radius,
                                                                        use_counts=self._use_counts,
                                                                        use_features=self._use_features)

    def calculate_score(self, molecules: List) -> ComponentSummary:
        query_fps = self._chemistry.mols_to_fingerprints(molecules, self._radius, self._use_counts, self._use_features)
        score = self._similarity.calculate_jaccard_distance(query_fps, self._ref_fingerprints)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters)
        return score_summary



from reinvent_scoring.scoring.score_components.standard.selectivity_component import SelectivityComponent
from reinvent_scoring.scoring.score_components.standard.custom_alerts_component import CustomAlerts
from reinvent_scoring.scoring.score_components.standard.jaccard_distance import JaccardDistance
from reinvent_scoring.scoring.score_components.standard.matching_substructure import MatchingSubstructure
from reinvent_scoring.scoring.score_components.standard.predictive_property_component import PredictivePropertyComponent
from reinvent_scoring.scoring.score_components.standard.qed_score import QedScore
from reinvent_scoring.scoring.score_components.standard.tanimoto_similarity import TanimotoSimilarity



from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.pip.base_pip_model_component import BasePiPModelComponent


class RatPKPiP(BasePiPModelComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._rat_pk_property = self.parameters.specific_parameters[self.component_specific_parameters.RAT_PK_PROPERTY]


    def _parse_single_compound(self, compound):
        return float(compound[self._rat_pk_property])



import glob
import os
import re
from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class BaseROCSComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def calculate_score_for_step(self, molecules: List, step=-1) -> ComponentSummary:
        return self.calculate_score(molecules, step)

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        # NOTE: valid_idxs are determined with RDKit not with Open Eye
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score = self._calculate_omega_score(valid_smiles, step)
        score = self._transformation_function(score, self.parameters.specific_parameters)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters)
        return score_summary

    @abstractmethod
    def _calculate_omega_score(self, smiles, step) -> np.array:
        raise NotImplementedError("_calculate_omega_score method is not implemented")



ROCS_DEFAULT_VALUES = dict(
    MAX_STEREO=0,
    MAX_CPUS=4,
    SAVE_ROCS_OVERLAYS=False,
    ENUM_STEREO=False,
    MAX_CONFS=200,
    EWINDOW=10,
    ROCS_OVERLAYS_DIR=None,
    ROCS_OVERLAYS_PREFIX="",
    PROTEIN_NEG_VOL_FILE=None,
    LIGAND_NEG_VOL_FILE=None,
    ROCS_INPUT=None,
    INPUT_TYPE=None,
    CUSTOM_CFF="",
    SHAPE_WEIGHT=None,
    COLOR_WEIGHT=None,
    NEGATIVE_VOLUME=False,
)


from typing import List

import numpy as np
from rdkit import Chem

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class MatchingSubstructure(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self.target_smarts = self.parameters.specific_parameters.get(self.component_specific_parameters.SMILES, [])
        self._validate_inputs(self.target_smarts)

    def calculate_score(self, molecules: List) -> ComponentSummary:
        score = self._substructure_match(molecules, self.target_smarts)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters)
        return score_summary

    def _substructure_match(self, query_mols, list_of_SMARTS):
        if len(list_of_SMARTS) == 0:
            return np.ones(len(query_mols), dtype=np.float32)

        match = [any([mol.HasSubstructMatch(Chem.MolFromSmarts(subst)) for subst in list_of_SMARTS
                      if Chem.MolFromSmarts(subst)]) for mol in query_mols]
        return 0.5 * (1 + np.array(match))

    def _validate_inputs(self, smiles):
        for smart in smiles:
            if Chem.MolFromSmarts(smart) is None:
                raise IOError(f"Invalid smarts pattern provided as a matching substructure: {smart}")





class ScoringFunctionComponentNameEnum:
    PARALLEL_ROCS_SIMILARITY = "parallel_rocs_similarity"
    SELECTIVITY = "selectivity"
    PREDICTIVE_PROPERTY = "predictive_property"
    ROCS_SIMILARITY = "rocs_similarity"
    MATCHING_SUBSTRUCTURE = "matching_substructure"
    TANIMOTO_SIMILARITY = "tanimoto_similarity"
    JACCARD_DISTANCE = "jaccard_distance"
    CUSTOM_ALERTS = "custom_alerts"
    QED_SCORE = "qed_score"
    MOLECULAR_WEIGHT = "molecular_weight"
    NUM_ROTATABLE_BONDS = "num_rotatable_bonds"
    NUM_HBD_LIPINSKI = "num_hbd_lipinski"
    NUM_HBA_LIPINSKI = "num_hba_lipinski"
    NUM_RINGS = "num_rings"
    TPSA = "tpsa"
    SLOGP = "slogp"
    GRAPH_LENGTH = "graph_length"
    NUMBER_OF_STEREO_CENTERS = "number_of_stereo_centers"
    TOTAL_SCORE = "total_score" # there is no actual component corresponding to this type
    REACTION_FILTERS = "reaction_filters"

    #NOTE: components below are AZ specific
    AZ_LOGD74 = "az_logd74"
    HLM_CLINT = "hlm_clint"
    RH_CLINT = "rh_clint"
    HH_CLINT = "hh_clint"
    SOLUBILITY_DD = "solubilityDD"
    CACO2_INTR = "caco2_intr"
    CACO2_EFFLUX = "caco2_efflux"
    HERG = "clab_herg"
    SA_SCORE = "sa_score"
    AZDOCK = "azdock"
    DOCKSTREAM = "dockstream"
    AZGARD = "azgard"
    AZ_LOGD74_PIP = "azlogd74"
    CACO2_INTR_PIP = "caco2-intrinsic-papp"
    CACO2_EFFLUX_PIP = "caco2-efflux"
    HH_CLINT_PIP = "hh-clint"
    HLM_CLINT_PIP = "hlm-clint"
    RH_CLINT_PIP = "rh-clint"
    SOLUBILITY_DD_PIP = "solubility-dd"
    HERG_PIP = "herg"
    KPUU_PIP = "rat-kpuu-brain"
    RAT_PK_PIP = "rat-pk"
    CLAB_TOP_20 = "clab_top_20"
    RA_SCORE = "rascore"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    def __setattr__(self, key, value):
        raise ValueError("Do not assign value to a ScoringFunctionComponentNameEnum field.")



import numpy as np
from rdkit.Chem.Descriptors import qed
from typing import List

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class QedScore(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def calculate_score(self, molecules: List) -> ComponentSummary:
        score = self._calculate_qed(molecules)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters)
        return score_summary

    def _calculate_qed(self, query_mols) -> np.array:
        qed_scores = []
        for mol in query_mols:
            try:
                qed_score = qed(mol)
            except ValueError:
                qed_score = 0.0
            qed_scores.append(qed_score)
        return np.array(qed_scores, dtype=np.float32)


from typing import List, Dict

import numpy as np

from reinvent_chemistry.descriptors import Descriptors
from reinvent_scoring.scoring.predictive_model.base_model_container import BaseModelContainer


class ScikitModelContainer(BaseModelContainer):
    def __init__(self, activity_model, model_type: str, specific_parameters: {}):
        """
        :type activity_model: scikit-learn type of model object
        :type model_type: can be "classification" or "regression"
        """
        self._activity_model = activity_model
        self._model_type = model_type
        self._molecules_to_descriptors = self._load_descriptor(specific_parameters)

    def predict(self, molecules: List, parameters: Dict) -> np.array:
        """
        Takes as input RDKit molecules and uses a pickled scikit-learn model to predict activities.
        :param molecules: This is a list of rdkit.Chem.Mol objects
        :param parameters: Those are descriptor-specific parameters.
        :return: numpy.array with activity predictions
        """
        return self.predict_from_mols(molecules, parameters)

    def predict_from_mols(self, molecules: List, parameters: dict):
        if len(molecules) == 0:
            return np.empty([])
        fps = self._molecules_to_descriptors(molecules, parameters)
        activity = self.predict_from_fingerprints(fps)
        return activity

    def predict_from_fingerprints(self, fps):
        if self._model_type == "regression":
            activity = self._activity_model.predict(fps)
        else:
            predictions = self._activity_model.predict_proba(fps)
            activity = predictions[:, 1]

        return activity

    def _load_descriptor(self, parameters: {}):
        descriptors = Descriptors()
        descriptor = descriptors.load_descriptor(parameters)
        return descriptor


import json
import os
from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.enums import EnvironmentalVariablesEnum
from reinvent_scoring.scoring.score_components.pip.base_rest_component import BaseRESTComponent
import requests
import time

class BasePiPModelBatchingComponent(BaseRESTComponent):
    def __init__(self, parameters: ComponentParameters):
        self._environment_keys = EnvironmentalVariablesEnum()
        super().__init__(parameters)
        self._get_header = self._create_get_header()

    def _execute_request(self, request_url, data, header) -> dict:
        request = requests.post(request_url, json=data, headers=header)
        if request.status_code != 202:
            raise ValueError(
                f" Status: {request.status_code} Reason: ({request.reason})."
                f"Response content: {request.content}\n"
                f"Response content: {request.text}"
            )
        response = self._get_results(request)

        return response.json()

    def _parse_response(self, response_json: dict, data_size: int) -> np.array:
        compounds = response_json['jsonData']['data']
        results_raw = np.empty(data_size, dtype=np.float32)
        results_raw[:] = np.nan

        try:
            for compound in compounds:
                try:
                    index = int(compound["id"])
                    results_raw[index] = self._parse_single_compound(compound)

                except (ValueError, TypeError, KeyError):
                    pass  # If parsing failed, keep value NaN for this compound and continue.
        finally:
            return results_raw

    @abstractmethod
    def _parse_single_compound(self, compound):
        raise NotImplementedError("_parse_compound method is not implemented")

    def _format_data(self, smiles: List[str]) -> dict:
        molecules = [{"molData": smi, "id": f"{i}"} for i, smi in enumerate(smiles)]
        data = {
            "jsonData": {
                "data": molecules,
                "metadata": {
                    "molFormat":
                        "smiles"
                },
                "parameters": {

                }
            }
        }
        return data

    def _create_url(self, async_path: str) -> str:
        pip_url = self._get_enviornment_variable(self._environment_keys.PIP_URL)
        request_url = pip_url.format(async_path)
        return request_url

    def _create_get_url(self, component_name) -> str:
        pip_url = self._get_enviornment_variable(self._environment_keys.PIP_GET_RESULTS)
        request_url = pip_url.format(component_name)
        return request_url

    def _create_header(self) -> dict:
        pip_key = self._get_enviornment_variable(self._environment_keys.PIP_KEY)

        header = {
            'Content-Type': 'application/vnd.az.batch.v1+json', 'x-api-key': pip_key,
            'Accept': 'application/vnd.az.resultset.v1+json',
            'Prefer': 'respond-async'
        }
        return header

    def _create_get_header(self) -> dict:
        pip_key = self._get_enviornment_variable(self._environment_keys.PIP_KEY)
        header = {'Content-Type': 'application/json', 'x-api-key': pip_key}
        return header

    def _get_enviornment_variable(self, variable: str) -> str:
        try:
            return os.environ[variable]
        except KeyError:
            return self._retrieve_pip_key_from_config(variable)

    def _retrieve_pip_key_from_config(self, variable: str) -> str:
        try:
            project_root = os.path.dirname(__file__)
            with open(os.path.join(project_root, '../../../configs/config.json'), 'r') as f:
                config = json.load(f)
            environmental_variables = config[self._environment_keys.ENVIRONMENTAL_VARIABLES]
            return environmental_variables[variable]
        except KeyError as ex:
            raise KeyError(f"Key {variable} not found in reinvent_scoring/configs/config.json")

    def _get_results(self, response):
        """
        Documentation for this get call can be found at:
        https://confluence.astrazeneca.net/pages/viewpage.action?spaceKey=ADD&title=Batching+and+caching+service
        """
        async_location = response.headers.get('Location', None)
        url = self._create_get_url(async_location)

        time_threshold = 300
        time_delay = 1

        while time_delay > 0 :

            try:
                response = requests.get(url=url, headers=self._get_header)
                time_delay = int(response.headers.get('retry-after', 0))
                time_threshold-=time_delay
                time.sleep(time_delay)
                if time_threshold <= 0:
                    raise TimeoutError(f'Terminated since retrieving results took too long')
            except:
                raise ValueError(
                    f" Status: {response.status_code} Reason: ({response.reason})."
                    f"Response content: {response.content}\n"
                    f"Response content: {response.text}"
                )
        result = response
        return result


from reinvent_scoring.scoring.diversity_filters.reinvent_core.identical_murcko_scaffold import IdenticalMurckoScaffold
from reinvent_scoring.scoring.diversity_filters.reinvent_core.no_scaffold_filter import NoScaffoldFilter
from reinvent_scoring.scoring.diversity_filters.reinvent_core.scaffold_similarity import ScaffoldSimilarity
from reinvent_scoring.scoring.diversity_filters.reinvent_core.identical_topological_scaffold import IdenticalTopologicalScaffold



import json
import subprocess
from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class BaseClabComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._clab_input_file = self.parameters.specific_parameters[self.component_specific_parameters.CLAB_INPUT_FILE]

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score, raw_score = self._calculate_score(valid_smiles)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)
        return score_summary

    def _calculate_score(self, smiles: List[str]) -> np.array:
        formatted = self._format_clab_input(smiles)
        self._write_clab_input(formatted)
        command = self._create_command(self._clab_input_file)
        results_string = self._send_request_get_stdout(command)
        results_raw = self._parse_multicompound_result(results_string, len(smiles))
        results = self._transformed_results(results_raw)
        results_raw = self._set_nan_to_zero(results_raw)
        return results, results_raw

    def _parse_multicompound_result(self, results: str, data_size: int) -> np.array:
        results_raw = np.empty(data_size, dtype=np.float32)
        results_raw[:] = np.nan
        try:
            loaded = json.loads(results)
            compounds = loaded[0]["Compounds"]
            for compound in compounds:
                try:
                    index = int(compound["Compound"])
                    results_raw[index] = self._parse_compound(compound)
                except (ValueError, TypeError, KeyError):
                    # ValueError: float("a")
                    # TypeError: 3["anything"]
                    # KeyError: compound["bad-name"]
                    pass  # If parsing failed, keep value NaN for this compound and continue.
        finally:
            return results_raw

    def _transformed_results(self, results_raw: np.array) -> np.array:
        """Returns np.array with non-NaN elements transformed by transformation function, and all NaN elements
        transformed into 0. """
        valid_mask = ~np.isnan(results_raw)
        results_raw_valid = results_raw[valid_mask]
        results_transformed = self._transformation_function(results_raw_valid, self.parameters.specific_parameters)
        results = np.zeros(len(results_raw), dtype=np.float32)
        results[valid_mask] = results_transformed
        return results

    def _set_nan_to_zero(self, results_raw: np.array) -> np.array:
        """Returns np.array with  all NaN elements transformed to 0. """
        valid_mask = ~np.isnan(results_raw)
        results_raw_valid = results_raw[valid_mask]
        results = np.zeros(len(results_raw), dtype=np.float32)
        results[valid_mask] = results_raw_valid
        return results

    @abstractmethod
    def _create_command(self, input_file) -> str:
        raise NotImplementedError("_create_command method is not implemented")

    @staticmethod
    def _send_request_get_stdout(command: str) -> str:
        with subprocess.Popen(command, stdout=subprocess.PIPE, shell=True, universal_newlines=True) as proc:
            stdout, stderr = proc.communicate()
        return stdout

    @abstractmethod
    def _parse_compound(self, compound):
        pass

    def _format_clab_input(self, smiles):
        """Prepares clab input compatible with clab_caco2_efflux, which requires quite strange json.

        Formats input to resemble the following for two molecules:
            {"smiles": "CCC AZ1\nCCCCC AZ2"}

        Alternative input is a simple SMILES (.smi) file, but it would require different extension (.smi).
        Sending simple SMILES-file with .json extension results in an error.
        Corresponding SMILES file example would be:
            CCC AZ1
            CCCCC AZ2
        """

        lines = [f"{smi} {i}" for i, smi in enumerate(smiles)]
        joined_lines = "\\n".join(lines)
        return f'{{"smiles": "{joined_lines}"}}'

    def _write_clab_input(self, data):
        with open(self._clab_input_file, 'w') as ofile:
            ofile.writelines(data)



from abc import ABC, abstractmethod
from typing import List

import numpy as np
from pathos.multiprocessing import ProcessPool

from reinvent_chemistry.conversions import Conversions
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.score_component_factory import ScoreComponentFactory
from reinvent_scoring.scoring.score_summary import ComponentSummary, FinalSummary
from reinvent_scoring.scoring.enums import ComponentSpecificParametersEnum
from reinvent_scoring.scoring.enums import ScoringFunctionComponentNameEnum


def _update_raw_score(summary: ComponentSummary, query_length: int, valid_indices: List[int]) -> ComponentSummary:
    if summary.raw_score is not None:
        raw_score = np.full(query_length, 0, dtype=np.float32)
        assert len(valid_indices) == len(summary.raw_score)
        for idx, value in zip(valid_indices, summary.raw_score):
            raw_score[idx] = value
        summary.raw_score = raw_score
    return summary


def _update_total_score(summary: ComponentSummary, query_length: int, valid_indices: List[int]) -> ComponentSummary:
    total_score = np.full(query_length, 0, dtype=np.float32)
    assert len(valid_indices) == len(summary.total_score)
    for idx, value in zip(valid_indices, summary.total_score):
        total_score[idx] = value
    summary.total_score = total_score
    summary = _update_raw_score(summary, query_length, valid_indices)
    return summary


def parallel_run(component_smiles_pair):
    component = component_smiles_pair[0]
    molecules = component_smiles_pair[1]
    valid_indices = component_smiles_pair[2]
    smiles = component_smiles_pair[3]
    scores = component.calculate_score(molecules)
    scores = _update_total_score(scores, len(smiles), valid_indices)
    return scores


def parallel_run_for_step(component_smiles_pair):
    component = component_smiles_pair[0]
    molecules = component_smiles_pair[1]
    valid_indices = component_smiles_pair[2]
    smiles = component_smiles_pair[3]
    step = component_smiles_pair[4]
    scores = component.calculate_score_for_step(molecules, step)
    scores = _update_total_score(scores, len(smiles), valid_indices)
    return scores


class BaseScoringFunction(ABC):
    def __init__(self, parameters: List[ComponentParameters], parallel=False):
        self.component_enum = ScoringFunctionComponentNameEnum()
        self.component_specific_parameters = ComponentSpecificParametersEnum()
        self._chemistry = Conversions()
        factory = ScoreComponentFactory(parameters)
        self.scoring_components = factory.create_score_components()
        if parallel:
            self.get_final_score = self._parallel_final_score
            self.get_final_score_for_step = self._parallel_final_score_for_step

    def get_final_score_for_step(self, smiles: List[str], step: int) -> FinalSummary:
        molecules, valid_indices = self._chemistry.smiles_to_mols_and_indices(smiles)
        query_size = len(smiles)
        summaries = [_update_total_score(sc.calculate_score_for_step(molecules, step), query_size, valid_indices) for sc
                     in self.scoring_components]
        return self._score_summary(summaries, smiles, valid_indices)

    def get_final_score(self, smiles: List[str]) -> FinalSummary:
        molecules, valid_indices = self._chemistry.smiles_to_mols_and_indices(smiles)
        query_size = len(smiles)
        summaries = [_update_total_score(sc.calculate_score(molecules), query_size, valid_indices) for sc
                     in self.scoring_components]
        return self._score_summary(summaries, smiles, valid_indices)

    def _score_summary(self, summaries: List[ComponentSummary], smiles: List[str],
                       valid_indices: List[int]) -> FinalSummary:

        penalty = self._compute_penalty_components(summaries, smiles)
        non_penlaty = self._compute_non_penalty_components(summaries, smiles)
        product = penalty * non_penlaty
        final_summary = self._create_final_summary(product, summaries, smiles, valid_indices)

        return final_summary

    def _create_final_summary(self, final_score, summaries: List[ComponentSummary], smiles: List[str],
                              valid_indices: List[int]) -> FinalSummary:

        return FinalSummary(total_score=np.array(final_score, dtype=np.float32),
                            scored_smiles=smiles,
                            valid_idxs=valid_indices,
                            scaffold_log_summary=summaries)

    def _compute_penalty_components(self, summaries: List[ComponentSummary], smiles: List[str]):
        penalty = np.full(len(smiles), 1, dtype=np.float32)

        for summary in summaries:
            if self._component_is_penalty(summary):
                penalty = penalty * summary.total_score

        return penalty

    @abstractmethod
    def _compute_non_penalty_components(self, summaries: List[ComponentSummary], smiles: List[str]):
        raise NotImplementedError("_score_summary method is not implemented")

    def _component_is_penalty(self, summary: ComponentSummary) -> bool:
        return (summary.parameters.component_type == self.component_enum.MATCHING_SUBSTRUCTURE) or (
                summary.parameters.component_type == self.component_enum.CUSTOM_ALERTS)

    def _parallel_final_score(self, smiles: List[str]) -> FinalSummary:
        molecules, valid_indices = self._chemistry.smiles_to_mols_and_indices(smiles)
        component_smiles_pairs = [[component, molecules, valid_indices, smiles] for component in
                                  self.scoring_components]
        pool = ProcessPool(nodes=len(self.scoring_components))
        mapped_pool = pool.map(parallel_run, component_smiles_pairs)
        pool.clear()
        return self._score_summary(mapped_pool, smiles, valid_indices)

    def _parallel_final_score_for_step(self, smiles: List[str], step: int) -> FinalSummary:
        molecules, valid_indices = self._chemistry.smiles_to_mols_and_indices(smiles)
        component_smiles_pairs = [[component, molecules, valid_indices, smiles, step] for component in
                                  self.scoring_components]
        pool = ProcessPool(nodes=len(self.scoring_components))
        mapped_pool = pool.map(parallel_run_for_step, component_smiles_pairs)
        pool.clear()
        return self._score_summary(mapped_pool, smiles, valid_indices)



from collections import namedtuple

import numpy as np
from openeye import oechem, oeomega, oeshape

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.rocs.base_rocs_component import BaseROCSComponent
from reinvent_scoring.scoring.enums import ROCSSimilarityMeasuresEnum
from reinvent_scoring.scoring.enums import ROCSInputFileTypesEnum
from reinvent_scoring.scoring.enums import ROCSSpecificParametersEnum


class RocsSimilarity(BaseROCSComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self.sim_measure_enum = ROCSSimilarityMeasuresEnum()
        self.input_types_enum = ROCSInputFileTypesEnum()
        self.param_names_enum = ROCSSpecificParametersEnum()
        self.shape_weight = self.parameters.specific_parameters[self.param_names_enum.SHAPE_WEIGHT]
        self.color_weight = self.parameters.specific_parameters[self.param_names_enum.COLOR_WEIGHT]
        self.sim_func_name_set = self.__get_similarity_name_set()
        cff_path = self.parameters.specific_parameters.get(self.param_names_enum.CUSTOM_CFF, None)
        self.prep = self.__set_prep(cff_path)
        self.overlay = self.__prepare_overlay(self.parameters.specific_parameters[self.param_names_enum.ROCS_INPUT],
                                              self.parameters.specific_parameters[self.param_names_enum.INPUT_TYPE])
        self.omega = self.__setup_omega()
        oechem.OEThrow.SetLevel(10000)

    def _calculate_omega_score(self, smiles, step=-1) -> np.array:
        scores = []
        predicate = getattr(oeshape, self.sim_func_name_set.predicate)()
        for smile in smiles:
            imol = oechem.OEMol()
            best_score = 0.0
            if oechem.OESmilesToMol(imol, smile):
                if self.omega(imol):
                    self.prep.Prep(imol)
                    score = oeshape.OEBestOverlayScore()
                    self.overlay.BestOverlay(score, imol, predicate)
                    best_score_shape = getattr(score, self.sim_func_name_set.shape)()
                    best_score_color = getattr(score, self.sim_func_name_set.color)()
                    best_score_color = correct_color_score(best_score_color)
                    best_score = ((self.shape_weight * best_score_shape) + (
                            self.color_weight * best_score_color)) / (self.shape_weight + self.color_weight)
            scores.append(best_score)
        return np.array(scores)

    def __setup_reference_molecule_with_shape_query(self, shape_query):
        qry = oeshape.OEShapeQuery()
        overlay = oeshape.OEOverlay()
        if oeshape.OEReadShapeQuery(shape_query, qry):
            overlay.SetupRef(qry)
        return overlay

    def __setup_reference_molecule(self, file_path):
        input_stream = oechem.oemolistream()
        input_stream.SetFormat(oechem.OEFormat_SDF)
        input_stream.SetConfTest(oechem.OEAbsoluteConfTest(compTitles=False))
        refmol = oechem.OEMol()
        if input_stream.open(file_path):
            oechem.OEReadMolecule(input_stream, refmol)
        cff = oeshape.OEColorForceField()
        if cff.Init(oeshape.OEColorFFType_ImplicitMillsDean):
            self.prep.SetColorForceField(cff)
        self.prep.Prep(refmol)
        overlay = oeshape.OEMultiRefOverlay()
        overlay.SetupRef(refmol)
        return overlay

    def __setup_omega(self):
        omegaOpts = oeomega.OEOmegaOptions()
        omegaOpts.SetStrictStereo(False)
        return oeomega.OEOmega(omegaOpts)

    def __get_similarity_name_set(self):
        similarity_collection_name = self.parameters.specific_parameters.get(self.param_names_enum.SIM_MEASURE,
                                                                        self.sim_measure_enum.TANIMOTO)
        similarity_collection = self.__similarity_collection(similarity_collection_name)
        return similarity_collection

    def __similarity_collection(self, sim_measure_type):
        _SIM_FUNC = namedtuple('sim_func', ['shape', 'color', 'predicate'])
        _SIM_DEF_DICT = {
            self.sim_measure_enum.TANIMOTO: _SIM_FUNC('GetTanimoto', 'GetColorTanimoto', 'OEHighestTanimotoCombo'),
            self.sim_measure_enum.REF_TVERSKY: _SIM_FUNC('GetRefTversky', 'GetRefColorTversky',
                                                         'OEHighestRefTverskyCombo'),
            self.sim_measure_enum.FIT_TVERSKY: _SIM_FUNC('GetFitTversky', 'GetFitColorTversky',
                                                         'OEHighestFitTverskyCombo'),
        }
        return _SIM_DEF_DICT.get(sim_measure_type)

    def __set_prep(self, cff_path):
        prep = oeshape.OEOverlapPrep()
        if cff_path is None:
            cff_path = oeshape.OEColorFFType_ImplicitMillsDean
        cff = oeshape.OEColorForceField()
        if cff.Init(cff_path):
            prep.SetColorForceField(cff)
        else:
            raise ValueError("Custom color force field initialisation failed")
        return prep

    def __prepare_overlay(self, file_path, overlay_type):
        overlays = {
            self.input_types_enum.SHAPE_QUERY: self.__setup_reference_molecule_with_shape_query,
            self.input_types_enum.SDF_QUERY: self.__setup_reference_molecule
        }
        overlay_function = overlays.get(overlay_type)
        overlay = overlay_function(file_path)
        return overlay

def correct_color_score(score):
    if score >= 1.0:
        score = 0.9 # or alternative
    return score


class DescriptorTypesEnum():
    _ECFP = "ecfp"
    _ECFP_COUNTS = "ecfp_counts"
    _MACCS_KEYS = "maccs_keys"
    _AVALON = "avalon"

    @property
    def ECFP(self):
        return self._ECFP

    @ECFP.setter
    def ECFP(self, value):
        raise ValueError("Do not assign value to a DescriptorTypesEnum field")

    @property
    def ECFP_COUNTS(self):
        return self._ECFP_COUNTS

    @ECFP_COUNTS.setter
    def ECFP_COUNTS(self, value):
        raise ValueError("Do not assign value to a DescriptorTypesEnum field")

    @property
    def MACCS_KEYS(self):
        return self._MACCS_KEYS

    @MACCS_KEYS.setter
    def MACCS_KEYS(self, value):
        raise ValueError("Do not assign value to a DescriptorTypesEnum field")

    @property
    def AVALON(self):
        return self._AVALON

    @AVALON.setter
    def AVALON(self, value):
        raise ValueError("Do not assign value to a DescriptorTypesEnum field")


import pickle

from typing import List

from reinvent_scoring.scoring.predictive_model.base_model_container import BaseModelContainer
from reinvent_scoring.scoring.predictive_model.model_container import ModelContainer
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary
from reinvent_scoring.scoring.score_transformations import TransformationFactory
from reinvent_scoring.scoring.enums import TransformationTypeEnum


class SelectivityComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._transformation_type = TransformationTypeEnum()
        self._model_transformation = self._assign_delta_transformation(
            {self.component_specific_parameters.TRANSFORMATION_TYPE: self._transformation_type.NO_TRANSFORMATION})
        self._activity_params = self._prepare_activity_parameters(parameters)
        self._off_target_params = self._prepare_offtarget_parameters(parameters)
        self._activity_model = self._load_model(self._activity_params)
        self._off_target_activity_model = self._load_model(self._off_target_params)
        self._delta_params = self._prepare_delta_parameters(parameters)
        self._delta_transformation = self._assign_delta_transformation(self._delta_params)

    def calculate_score(self, molecules: List) -> ComponentSummary:
        score, offtarget_score = self._calculate_offtarget_activity(molecules, self._activity_params,
                                                                    self._off_target_params, self._delta_params)
        score_summary = ComponentSummary(total_score=score, parameters=self._off_target_params,
                                         raw_score=offtarget_score)
        return score_summary

    def _load_model(self, parameters: ComponentParameters):
        try:
            activity_model = self._load_scikit_model(parameters)
        except Exception as e:
            model_path = self.parameters.specific_parameters.get(self.component_specific_parameters.MODEL_PATH, "")
            raise Exception(f"The loaded file `{model_path}` isn't a valid scikit-learn model: {e}.")
        return activity_model

    def _load_scikit_model(self, parameters: ComponentParameters) -> BaseModelContainer:
        model_path = parameters.specific_parameters.get(self.component_specific_parameters.MODEL_PATH, "")
        with open(model_path, "rb") as f:
            scikit_model = pickle.load(f)

            models_are_identical = self._activity_params.specific_parameters[
                                       self.component_specific_parameters.SCIKIT] == \
                                   self._off_target_params.specific_parameters[
                                       self.component_specific_parameters.SCIKIT]

            model_is_regression = self._off_target_params.specific_parameters[
                                      self.component_specific_parameters.SCIKIT] == "regression"

            both_models_are_regression = models_are_identical and model_is_regression

            if both_models_are_regression:
                parameters.specific_parameters[self.component_specific_parameters.TRANSFORMATION] = False

            self._assign_model_transformation(both_models_are_regression)

            packaged_model = ModelContainer(scikit_model, parameters.specific_parameters)
        return packaged_model

    def _calculate_offtarget_activity(self, molecules, activity_params, offtarget_params, delta_params):
        raw_activity_score = self._activity_model\
            .predict(molecules, activity_params.specific_parameters)
        raw_offtarget_score = self._off_target_activity_model \
            .predict(molecules, offtarget_params.specific_parameters)
        activity_score = self._apply_model_transformation(raw_activity_score, activity_params.specific_parameters)
        offtarget_score = self._apply_model_transformation(raw_offtarget_score, offtarget_params.specific_parameters)

        delta = activity_score - offtarget_score

        transformed_score = self._delta_transformation(delta, delta_params) if delta_params[
            self.component_specific_parameters.TRANSFORMATION] else delta
        transformed_score[transformed_score < 0.01] = 0.01

        return transformed_score, raw_offtarget_score

    def _assign_delta_transformation(self, specific_parameters: {}):
        factory = TransformationFactory()
        transform_function = factory.get_transformation_function(specific_parameters)
        return transform_function

    def _prepare_activity_parameters(self, parameters: ComponentParameters) -> ComponentParameters:
        model_path = parameters.specific_parameters["activity_model_path"]
        specific_parameters = parameters.specific_parameters["activity_specific_parameters"]
        specific_parameters['model_path'] = model_path
        activity_params = ComponentParameters(name=self.parameters.name,
                                              weight=self.parameters.weight,
                                              component_type=self.parameters.component_type,
                                              specific_parameters=specific_parameters
                                              )
        return activity_params

    def _prepare_offtarget_parameters(self, parameters: ComponentParameters) -> ComponentParameters:
        model_path = parameters.specific_parameters["offtarget_model_path"]
        specific_parameters = parameters.specific_parameters["offtarget_specific_parameters"]
        specific_parameters['model_path'] = model_path
        offtarget_params = ComponentParameters(name=self.parameters.name,
                                               weight=self.parameters.weight,
                                               component_type=self.parameters.component_type,
                                               specific_parameters=specific_parameters
                                               )
        return offtarget_params

    def _prepare_delta_parameters(self, parameters: ComponentParameters) -> dict:
        specific_params = parameters.specific_parameters["delta_transformation_parameters"]
        specific_params[self.component_specific_parameters.TRANSFORMATION] = \
            "regression" == self._activity_params.specific_parameters[self.component_specific_parameters.SCIKIT] == \
            self._off_target_params.specific_parameters[self.component_specific_parameters.SCIKIT]
        return specific_params

    def _apply_model_transformation(self, predicted_activity, parameters: dict):
        if parameters.get(self.component_specific_parameters.TRANSFORMATION, False):
            activity = self._model_transformation(predicted_activity, parameters)
        else:
            activity = predicted_activity
        return activity

    def _assign_model_transformation(self, both_models_are_regression: bool):
        if both_models_are_regression:
            return
        if self._activity_params.specific_parameters.get(self.component_specific_parameters.SCIKIT) == "regression":
            self._model_transformation = self._assign_delta_transformation(self._activity_params.specific_parameters)
        if self._off_target_params.specific_parameters.get(self.component_specific_parameters.SCIKIT) == "regression":
            self._model_transformation = self._assign_delta_transformation(self._off_target_params.specific_parameters)



import numpy as np
import math
from scipy.interpolate import interp1d

from reinvent_scoring.scoring.enums import ComponentSpecificParametersEnum
from reinvent_scoring.scoring.enums import TransformationTypeEnum


class TransformationFactory:

    def __init__(self):
        self._csp_enum = ComponentSpecificParametersEnum()
        self._transformation_function_registry = self._default_transformation_function_registry()

    def _default_transformation_function_registry(self) -> dict:
        enum = TransformationTypeEnum()
        transformation_list = {
            enum.SIGMOID: self.sigmoid_transformation,
            enum.REVERSE_SIGMOID: self.reverse_sigmoid_transformation,
            enum.DOUBLE_SIGMOID: self.double_sigmoid,
            enum.NO_TRANSFORMATION: self.no_transformation,
            enum.RIGHT_STEP: self.right_step,
            enum.LEFT_STEP: self.left_step,
            enum.STEP: self.step,
            enum.CUSTOM_INTERPOLATION: self.custom_interpolation
        }
        return transformation_list

    def get_transformation_function(self, parameters: dict):
        transformation_type = parameters[self._csp_enum.TRANSFORMATION_TYPE]
        transformation_function = self._transformation_function_registry[transformation_type]
        return transformation_function

    def no_transformation(self, predictions: list, parameters: dict) -> np.array:
        return np.array(predictions, dtype=np.float32)

    def right_step(self, predictions, parameters) -> np.array:
        _low = parameters[self._csp_enum.LOW]

        def _right_step_formula(value, low):
            if value >= low:
                return 1
            return 0

        transformed = [_right_step_formula(value, _low) for value in predictions]
        return np.array(transformed, dtype=np.float32)

    def left_step(self, predictions, parameters) -> np.array:
        _low = parameters[self._csp_enum.LOW]

        def _left_step_formula(value, low):
            if value <= low:
                return 1
            return 0

        transformed = [_left_step_formula(value, _low) for value in predictions]
        return np.array(transformed, dtype=np.float32)

    def step(self, predictions, parameters) -> np.array:
        _low = parameters[self._csp_enum.LOW]
        _high = parameters[self._csp_enum.HIGH]

        def _step_formula(value, low, high):
            if low <= value <= high:
                return 1
            return 0

        transformed = [_step_formula(value, _low, _high) for value in predictions]
        return np.array(transformed, dtype=np.float32)

    def sigmoid_transformation(self, predictions: list, parameters: dict) -> np.array:
        _low = parameters[self._csp_enum.LOW]
        _high = parameters[self._csp_enum.HIGH]
        _k = parameters[self._csp_enum.K]

        def _exp(pred_val, low, high, k) -> float:
            return math.pow(10, (10 * k * (pred_val - (low + high) * 0.5) / (low - high)))

        transformed = [1 / (1 + _exp(pred_val, _low, _high, _k)) for pred_val in predictions]
        return np.array(transformed, dtype=np.float32)

    def reverse_sigmoid_transformation(self, predictions: list, parameters: dict) -> np.array:
        _low = parameters[self._csp_enum.LOW]
        _high = parameters[self._csp_enum.HIGH]
        _k = parameters[self._csp_enum.K]

        def _reverse_sigmoid_formula(value, low, high, k) -> float:
            try:
                return 1 / (1 + 10 ** (k * (value - (high + low) / 2) * 10 / (high - low)))
            except:
                return 0

        transformed = [_reverse_sigmoid_formula(pred_val, _low, _high, _k) for pred_val in predictions]
        return np.array(transformed, dtype=np.float32)

    def double_sigmoid(self, predictions: list, parameters: dict) -> np.array:
        _low = parameters[self._csp_enum.LOW]
        _high = parameters[self._csp_enum.HIGH]
        _coef_div = parameters[self._csp_enum.COEF_DIV]
        _coef_si = parameters[self._csp_enum.COEF_SI]
        _coef_se = parameters[self._csp_enum.COEF_SE]

        def _double_sigmoid_formula(value, low, high, coef_div=100., coef_si=150., coef_se=150.):
            try:
                A = 10 ** (coef_se * (value / coef_div))
                B = (10 ** (coef_se * (value / coef_div)) + 10 ** (coef_se * (low / coef_div)))
                C = (10 ** (coef_si * (value / coef_div)) / (
                        10 ** (coef_si * (value / coef_div)) + 10 ** (coef_si * (high / coef_div))))
                return (A / B) - C
            except:
                return 0

        transformed = [_double_sigmoid_formula(pred_val, _low, _high, _coef_div, _coef_si, _coef_se) for pred_val in
                       predictions]
        return np.array(transformed, dtype=np.float32)

    def custom_interpolation(self, predictions: list, parameters: dict) -> np.array:
        """Adapted from the paper:
        'Efficient Multi-Objective Molecular Optimization in a Continuous Latent Space'
        by Robin Winter, Floriane Montanari, Andreas Steffen, Hans Briem, Frank Noé and Djork-Arné Clevert.
        """

        def _transformation_function(interpolation_map, truncate_left=True, truncate_right=True):
            origin = [point['origin'] for point in interpolation_map]
            destination = [point['destination'] for point in interpolation_map]
            assert len(origin) == len(destination)

            if truncate_left:
                origin = [origin[0] - 1] + origin
                destination = [destination[0]] + destination
            if truncate_right:
                origin.append(origin[-1] + 1)
                destination.append(destination[-1])
            return interp1d(origin, destination, fill_value='extrapolate')

        desirability = parameters.get(self._csp_enum.INTERPOLATION_MAP, [{"origin": 0.0, "destination": 0.0},
                                                                         {"origin": 1.0, "destination": 1.0}])
        truncate_left = parameters.get(self._csp_enum.TRUNCATE_LEFT, True)
        truncate_right = parameters.get(self._csp_enum.TRUNCATE_RIGHT, True)

        transformation = _transformation_function(desirability, truncate_left, truncate_right)
        transformed = transformation(predictions)

        return transformed



class ROCSSimilarityMeasuresEnum():
    _TANIMOTO = "Tanimoto"
    _REF_TVERSKY = "RefTversky"
    _FIT_TVERSKY = "FitTversky"

    # Need to stick to this pattern of getter/setter as the alternative more concise one cannot
    # be pickled for multiprocessing

    @property
    def TANIMOTO(self):
        return self._TANIMOTO

    @TANIMOTO.setter
    def TANIMOTO(self, value):
        raise ValueError("Do not assign value to a ROCSSimiliarityMeasuresEnum field")

    @property
    def REF_TVERSKY(self):
        return self._REF_TVERSKY

    @REF_TVERSKY.setter
    def REF_TVERSKY(self, value):
        raise ValueError("Do not assign value to a ROCSSimiliarityMeasuresEnum field")

    @property
    def FIT_TVERSKY(self):
        return self._FIT_TVERSKY

    @FIT_TVERSKY.setter
    def FIT_TVERSKY(self, value):
        raise ValueError("Do not assign value to a ROCSSimiliarityMeasuresEnum field")



from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.function.base_scoring_function import BaseScoringFunction
from reinvent_scoring.scoring.score_summary import ComponentSummary


class CustomSum(BaseScoringFunction):

    def __init__(self, parameters: List[ComponentParameters], parallel=False):
        super().__init__(parameters, parallel)

    def _compute_non_penalty_components(self, summaries: List[ComponentSummary], smiles: List[str]):
        total_sum = np.full(len(smiles), 0, dtype=np.float32)
        all_weights = 0.

        for summary in summaries:
            if not self._component_is_penalty(summary):
                total_sum = total_sum + summary.total_score * summary.parameters.weight
                all_weights += summary.parameters.weight

        if all_weights == 0:
            """There are no non-penalty components and return array of ones. 
            This is needed so that it can work in cases where only penalty components are used"""
            return np.full(len(smiles), 1, dtype=np.float32)

        return total_sum / all_weights



from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring import CustomProduct, CustomSum
from reinvent_scoring.scoring.function.base_scoring_function import BaseScoringFunction
from reinvent_scoring.scoring.scoring_function_parameters import ScoringFunctionParameters
from reinvent_scoring.scoring.enums import ScoringFunctionNameEnum


class ScoringFunctionFactory:

    def __new__(cls, sf_parameters: ScoringFunctionParameters) -> BaseScoringFunction:
        enum = ScoringFunctionNameEnum()
        scoring_function_registry = {
            enum.CUSTOM_PRODUCT: CustomProduct,
            enum.CUSTOM_SUM: CustomSum
        }
        return cls.create_scoring_function_instance(sf_parameters, scoring_function_registry)

    @staticmethod
    def create_scoring_function_instance(sf_parameters: ScoringFunctionParameters,
                                         scoring_function_registry: dict) -> BaseScoringFunction:
        """Returns a scoring function instance"""
        scoring_function = scoring_function_registry[sf_parameters.name]
        parameters = [ComponentParameters(**p) for p in sf_parameters.parameters]

        return scoring_function(parameters, sf_parameters.parallel)



class ROCSSpecificParametersEnum():
    _ROCS_INPUT = "rocs_input"
    _INPUT_TYPE = "input_type"
    _SHAPE_WEIGHT = "shape_weight"
    _COLOR_WEIGHT = "color_weight"
    _SIM_MEASURE = "similarity_measure"
    _MAX_CPUS = "max_num_cpus"
    _CUSTOM_CFF = "custom_cff"
    _SAVE_ROCS_OVERLAYS = "save_rocs_overlays"
    _ROCS_OVERLAYS_DIR = "rocs_overlays_dir"
    _ROCS_OVERLAYS_PREFIX = "rocs_overlays_prefix"
    _ENUM_STEREO = "enumerate_stereo"
    _MAX_STEREO = "max_stereocenters"
    _NEGATIVE_VOLUME = "negative_volume"
    _PROTEIN_NEG_VOL_FILE = "protein_neg_vol_file"
    _LIGAND_NEG_VOL_FILE = "ligand_neg_vol_file"
    _MAX_CONFS = "max_confs"
    _EWINDOW = "ewindow"

    # Need to stick to this pattern of getter/setter as the alternative more concise one cannot
    # be pickled for multiprocessing

    @property
    def ROCS_INPUT(self):
        return self._ROCS_INPUT

    @ROCS_INPUT.setter
    def ROCS_INPUT(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def INPUT_TYPE(self):
        return self._INPUT_TYPE

    @INPUT_TYPE.setter
    def INPUT_TYPE(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def SHAPE_WEIGHT(self):
        return self._SHAPE_WEIGHT

    @SHAPE_WEIGHT.setter
    def SHAPE_WEIGHT(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def COLOR_WEIGHT(self):
        return self._COLOR_WEIGHT

    @COLOR_WEIGHT.setter
    def COLOR_WEIGHT(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def SIM_MEASURE(self):
        return self._SIM_MEASURE

    @SIM_MEASURE.setter
    def SIM_MEASURE(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def MAX_CPUS(self):
        return self._MAX_CPUS

    @MAX_CPUS.setter
    def MAX_CPUS(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def CUSTOM_CFF(self):
        return self._CUSTOM_CFF

    @CUSTOM_CFF.setter
    def CUSTOM_CFF(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def SAVE_ROCS_OVERLAYS(self):
        return self._SAVE_ROCS_OVERLAYS

    @SAVE_ROCS_OVERLAYS.setter
    def SAVE_ROCS_OVERLAYS(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def ROCS_OVERLAYS_DIR(self):
        return self._ROCS_OVERLAYS_DIR

    @ROCS_OVERLAYS_DIR.setter
    def ROCS_OVERLAYS_DIR(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def ROCS_OVERLAYS_PREFIX(self):
        return self._ROCS_OVERLAYS_PREFIX

    @ROCS_OVERLAYS_PREFIX.setter
    def ROCS_OVERLAYS_PREFIX(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def ENUM_STEREO(self):
        return self._ENUM_STEREO

    @ENUM_STEREO.setter
    def ENUM_STEREO(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def MAX_STEREO(self):
        return self._MAX_STEREO

    @MAX_STEREO.setter
    def MAX_STEREO(self, value):
        raise ValueError("Do not assign value to a ROCSSpecificParametersEnum field")

    @property
    def PROTEIN_NEG_VOL_FILE(self):
        return self._PROTEIN_NEG_VOL_FILE

    @PROTEIN_NEG_VOL_FILE.setter
    def PROTEIN_NEG_VOL_FILE(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")

    @property
    def NEGATIVE_VOLUME(self):
        return self._NEGATIVE_VOLUME

    @NEGATIVE_VOLUME.setter
    def NEGATIVE_VOLUME(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")

    @property
    def LIGAND_NEG_VOL_FILE(self):
        return self._LIGAND_NEG_VOL_FILE

    @LIGAND_NEG_VOL_FILE.setter
    def LIGAND_NEG_VOL_FILE(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")

    @property
    def MAX_CONFS(self):
        return self._MAX_CONFS

    @MAX_CONFS.setter
    def MAX_CONFS(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")

    @property
    def EWINDOW(self):
        return self._EWINDOW

    @EWINDOW.setter
    def EWINDOW(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")


from typing import List, Dict, Any

import numpy as np
from reinvent_chemistry.conversions import Conversions

from reinvent_scoring.scoring.predictive_model.base_model_container import BaseModelContainer


class OptunaModelContainer(BaseModelContainer):
    def __init__(self, activity_model):
        """
        :type activity_model: scikit-learn object
        """
        self._activity_model = activity_model
        self._conversions = Conversions()

    def predict(self, molecules: List[Any], parameters: Dict) -> np.array:
        """
        Takes a list of smiles as input an predicts activities.
        :param molecules:
        :param parameters:
        :return:
        """

        if len(molecules) == 0:
            return np.empty([])

        smiles = [self._conversions.mol_to_smiles(mol) for mol in molecules]
        activity = self._activity_model.predict_from_smiles(smiles)

        return activity





from copy import deepcopy
from typing import List

import numpy as np

from reinvent_scoring.scoring.diversity_filters.lib_invent.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.lib_invent.diversity_filter_parameters import DiversityFilterParameters
# The import below is a deal breaker
# from reinvent_scoring.scoring.score_summary import FinalSummary



class NoFilter(BaseDiversityFilter):
    """Doesn't penalize compounds."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)

    def update_score(self, score_summary, sampled_sequences: List, step=0) -> np.array:
        score_summary = deepcopy(score_summary)
        scores = score_summary.total_score
        for i in score_summary.valid_idxs:
            if scores[i] >= self.parameters.minscore:
                smile = score_summary.scored_smiles[i]
                decorations = f'{sampled_sequences[i].scaffold}|{sampled_sequences[i].decoration}'
                self._add_to_memory(i, scores[i], smile, decorations, score_summary.scaffold_log, step)
        return scores



class DiversityFilterEnum():
    IDENTICAL_TOPOLOGICAL_SCAFFOLD = "IdenticalTopologicalScaffold"
    IDENTICAL_MURCKO_SCAFFOLD = "IdenticalMurckoScaffold"
    SCAFFOLD_SIMILARITY = "ScaffoldSimilarity"
    NO_FILTER = "NoFilter"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    def __setattr__(self, key, value):
        raise ValueError("Do not assign value to a ScaffoldFilterEnum field.")



class LoggingModeEnum:
    LOCAL = "local"
    REMOTE = "remote"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    def __setattr__(self, key, value):
        raise ValueError("Do not assign value to a LoggingModeEnum field.")




import math
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.function.base_scoring_function import BaseScoringFunction
from reinvent_scoring.scoring.score_summary import ComponentSummary


class CustomProduct(BaseScoringFunction):

    def __init__(self, parameters: List[ComponentParameters], parallel=False):
        super().__init__(parameters, parallel)

    def _calculate_pow(self, values, weight):
        y = [math.pow(value, weight) for value in values]
        return np.array(y, dtype=np.float32)

    def _get_all_weights(self, summaries: List[ComponentSummary]) -> int:
        all_weights = []

        for summary in summaries:
            if not self._component_is_penalty(summary):
                all_weights.append(summary.parameters.weight)
        return sum(all_weights)

    def _compute_non_penalty_components(self, summaries: List[ComponentSummary], smiles: List[str]):
        product = np.full(len(smiles), 1, dtype=np.float32)
        all_weights = self._get_all_weights(summaries)

        for summary in summaries:
            if not self._component_is_penalty(summary):
                comp_pow = self._calculate_pow(summary.total_score, summary.parameters.weight / all_weights)
                product = product * comp_pow

        return product



from rdkit import Chem
from typing import List

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class CustomAlerts(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self.custom_alerts = self.parameters.specific_parameters.get(self.component_specific_parameters.SMILES, [''])

    def calculate_score(self, molecules: List) -> ComponentSummary:
        score = self._substructure_match(molecules, self.custom_alerts)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters)
        return score_summary

    def _substructure_match(self, query_mols, list_of_SMARTS):
        match = [any([mol.HasSubstructMatch(Chem.MolFromSmarts(subst)) for subst in list_of_SMARTS
                      if Chem.MolFromSmarts(subst)]) for mol in query_mols]
        reverse = [1 - m for m in match]
        return reverse





class ScoringFunctionNameEnum:
    CUSTOM_PRODUCT = "custom_product"
    CUSTOM_SUM = "custom_sum"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    def __setattr__(self, key, value):
        raise ValueError("Do not assign value to a ScoringFunctionNameEnum field.")



from typing import List

import pandas as pd
from reinvent_scoring.scoring.score_summary import ComponentSummary
from reinvent_scoring.scoring.enums.scoring_function_component_enum import ScoringFunctionComponentNameEnum


class DiversityFilterMemory:

    def __init__(self):
        self._sf_component_name = ScoringFunctionComponentNameEnum()
        df_dict = {"Step": [], "Scaffold": [], "SMILES": []}
        self._memory_dataframe = pd.DataFrame(df_dict)

    def update(self, indx: int, score: float, smile: str, scaffold: str, components: List, step: int):
        component_scores = {c.parameters.name: float(c.total_score[indx]) for c in components}
        component_scores = self._include_raw_score(indx, component_scores, components)
        component_scores[self._sf_component_name.TOTAL_SCORE] = float(score)
        if not self.smiles_exists(smile):
            self._add_to_memory_dataframe(step, smile, scaffold, component_scores)

    def _add_to_memory_dataframe(self, step: int, smile: str, scaffold: str, component_scores: {} = None):
        data = []
        headers = []
        for name, score in component_scores.items():
            headers.append(name)
            data.append(score)
        headers.append("Step")
        data.append(step)
        headers.append("Scaffold")
        data.append(scaffold)
        headers.append("SMILES")
        data.append(smile)
        new_data = pd.DataFrame([data], columns=headers)
        self._memory_dataframe = pd.concat([self._memory_dataframe, new_data], ignore_index=True, sort=False)

    def get_memory(self) -> pd.DataFrame:
        return self._memory_dataframe

    def set_memory(self, memory: pd.DataFrame):
        self._memory_dataframe = memory

    def smiles_exists(self, smiles: str):
        if len(self._memory_dataframe) == 0:
            return False
        return smiles in self._memory_dataframe['SMILES'].values

    def scaffold_instances_count(self, scaffold: str):
        return (self._memory_dataframe["Scaffold"].values == scaffold).sum()

    def number_of_scaffolds(self):
        return len(set(self._memory_dataframe["Scaffold"].values))

    def number_of_smiles(self):
        return len(set(self._memory_dataframe["SMILES"].values))

    def _include_raw_score(self, indx: int, component_scores: dict, components: List[ComponentSummary]):
        raw_scores = {f'raw_{c.parameters.name}': float(c.raw_score[indx]) for c in components if
                      c.raw_score is not None}
        all_scores = {**component_scores, **raw_scores}
        return all_scores



from reinvent_scoring.scoring.score_components.base_score_component import *
from reinvent_scoring.scoring.score_components.synthetic_accessibility import *
from reinvent_scoring.scoring.score_components.physchem import *
from reinvent_scoring.scoring.score_components.pip import *
from reinvent_scoring.scoring.score_components.structural import *
from reinvent_scoring.scoring.score_components.physchem import *
from reinvent_scoring.scoring.score_components.standard import *
from reinvent_scoring.scoring.score_components.clab import *
from reinvent_scoring.scoring.score_components.rocs import *



from abc import abstractmethod
from typing import List

import numpy as np
from reinvent_chemistry.phys_chem_descriptors import PhysChemDescriptors

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class BasePhysChemComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._phys_chem_descriptors = PhysChemDescriptors()

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        score, raw_score = self._calculate_score(molecules)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)
        return score_summary

    def _calculate_score(self, query_mols) -> np.array:
        scores = []
        for mol in query_mols:
            try:
                score = self._calculate_phys_chem_property(mol)
            except ValueError:
                score = 0.0
            scores.append(score)
        transformed_scores = self._transformation_function(scores, self.parameters.specific_parameters)
        return np.array(transformed_scores, dtype=np.float32), np.array(scores, dtype=np.float32)

    @abstractmethod
    def _calculate_phys_chem_property(self, mol):
        raise NotImplementedError("_calculate_phys_chem_property method is not implemented")




from reinvent_scoring.scoring.score_components.physchem.number_of_stereo_centers import NumberOfStereoCenters
from reinvent_scoring.scoring.score_components.physchem.tpsa import PSA
from reinvent_scoring.scoring.score_components.physchem.hba import HBA_Lipinski
from reinvent_scoring.scoring.score_components.physchem.hbd import HBD_Lipinski
from reinvent_scoring.scoring.score_components.physchem.mol_weight import MolWeight
from reinvent_scoring.scoring.score_components.physchem.num_rings import NumRings
from reinvent_scoring.scoring.score_components.physchem.rot_bonds import RotatableBonds
from reinvent_scoring.scoring.score_components.physchem.slogp import SlogP
from reinvent_scoring.scoring.score_components.physchem.graph_length import GraphLength



from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.clab.base_clab_component import BaseClabComponent


class HERG(BaseClabComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def _create_command(self, input_file):
        name = "clab_herg"
        command = f"ml {name}; {name} {input_file}"
        return command

    def _parse_compound(self, compound):
        return float(compound["Prediction"]["hERG_IC50"])



from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.physchem.base_physchem_component import BasePhysChemComponent


class HBA_Lipinski(BasePhysChemComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def _calculate_phys_chem_property(self, mol):
        return self._phys_chem_descriptors.hba_libinski(mol)



from reinvent_scoring.scoring.function import *
from reinvent_scoring.scoring.enums import *
from reinvent_scoring.scoring.score_components import *

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_summary import FinalSummary, ComponentSummary, LoggableComponent
from reinvent_scoring.scoring.score_transformations import TransformationFactory
from reinvent_scoring.scoring.scoring_function_factory import ScoringFunctionFactory
from reinvent_scoring.scoring.scoring_function_parameters import ScoringFunctionParameters



import io
import json
import os
import shutil
import subprocess
import tempfile
import time

import numpy as np
from typing import List, Tuple

from reinvent_scoring.scoring.utils import _is_development_environment

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.console_invoked.base_console_invoked_component import BaseConsoleInvokedComponent


class AZgard(BaseConsoleInvokedComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._configuration_path = self.parameters.specific_parameters[self.component_specific_parameters.AZGARD_CONFPATH]
        self._executor_script_path = self.parameters.specific_parameters[self.component_specific_parameters.AZGARD_EXECUTORSCRIPTPATH]
        self._environment_path = self.parameters.specific_parameters[self.component_specific_parameters.AZGARD_ENVPATH]
        self._values_key = self.parameters.specific_parameters[self.component_specific_parameters.AZGARD_VALUES_KEY]

    def _add_debug_mode_if_selected(self, command):
        if self.parameters.specific_parameters.get(self.component_specific_parameters.AZGARD_DEBUG, False)\
                or _is_development_environment():
            command = ' '.join([command, "-debug"])
        return command

    def _create_command(self, step, input_json_path: str, output_json_path: str):
        # use "step" as well for the write-out
        global_variables = "".join(["\"input_json_path:", input_json_path, "\" ",
                                    "\"output_json_path:", output_json_path, "\" ",
                                    "\"step_id:", str(step), "\""])
        command = ' '.join([self._environment_path,
                            self._executor_script_path,
                            "-conf", self._configuration_path,
                            "--global_variables", global_variables])

        # check, if AZgard is to be executed in debug mode, which will cause its loggers to print out
        # much more detailed information
        command = self._add_debug_mode_if_selected(command)
        return command

    def _prepare_input_data_JSON(self, path: str, smiles: List[str]):
        """Needs to look something like:
           {
               "names": ["0", "1", "3"],
               "smiles": ["C#CCCCn1...", "CCCCn1c...", "CC(C)(C)CCC1(c2..."]
           }"""
        names = [str(idx) for idx in range(len(smiles))]
        input_dict = {"names": names,
                      "smiles": smiles}
        with open(path, 'w') as f:
            json.dump(input_dict, f, indent=4)

    def _select_values(self, data: dict) -> list:
        for value_dict in data["results"]:
            if self._values_key == value_dict["value_key"]:
                return value_dict["values"]
        return []

    def _parse_output_data_json(self, path: str) -> Tuple[List[str], List[float]]:
        """Needs to look something like:
           {
               "results": [{
                   "value_key": "docking_score",
                   "values": ["-5.88841", "-5.72676", "-7.30167"]},
                           {
                   "value_key": "shape_similarity",
                   "values": ["0.476677", "0.458017", "0.510676"]},
                           {
                   "value_key": "esp_similarity",
                   "values": ["0.107989", "0.119446", "0.100109"]}],
               "names": ["0", "1", "2"]
           }"""
        names_list = []
        values_list = []

        if not os.path.isfile(path):
            raise FileNotFoundError(f"Output file {path} does not exist, indicating that execution of AZgard failed entirely. Check your setup and the log file.")

        with open(path, 'r') as f:
            data = f.read().replace("\r", "").replace("\n", "")
        data = json.loads(data)
        raw_values_list = self._select_values(data=data)

        for idx in range(len(data["names"])):
            names_list.append(data["names"][idx])
            try:
                score = float(raw_values_list[idx])
            except ValueError:
                score = 0
            values_list.append(score)

        return names_list, values_list

    def _execute_command(self, command: str, final_file_path: str = None):
        # execute the pre-defined command
        with subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              shell=True) as proc:
            wrapped_proc_in = io.TextIOWrapper(proc.stdin, "utf-8")
            wrapped_proc_out = io.TextIOWrapper(proc.stdout, "utf-8")
            wrapped_proc_in.close()
            wrapped_proc_out.close()
            proc.wait()
            proc.terminate()

        # wait in case the final file has not been written yet or is empty (in case of a filesystem delay / hick-up)
        if final_file_path is not None:
            for _ in range(5):
                if os.path.isfile(final_file_path) and os.path.getsize(final_file_path) > 0:
                    break
                else:
                    time.sleep(3)

    def _calculate_score(self, smiles: List[str], step) -> np.array:
        # make temporary folder and set input and output paths
        tmp_dir = tempfile.mkdtemp()
        input_json_path = os.path.join(tmp_dir, "input.json")
        output_json_path = os.path.join(tmp_dir, "output.json")

        # save the smiles in an AZgard compatible JSON file
        self._prepare_input_data_JSON(path=input_json_path, smiles=smiles)

        # create the external command
        command = self._create_command(step=step,
                                       input_json_path=input_json_path,
                                       output_json_path=output_json_path)

        # execute the AZgard component
        self._execute_command(command=command, final_file_path=output_json_path)

        # TODO: smiles_ids are not used yet; will be important to match the return values back
        # parse the output
        smiles_ids, scores = self._parse_output_data_json(path=output_json_path)

        # apply transformation
        transformed_scores = self._transformation_function(scores, self.parameters.specific_parameters)

        # clean up
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)

        return np.array(transformed_scores), np.array(scores)



import numpy as np
from typing import List

from reinvent_scoring.scoring.utils import _is_development_environment

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.structural.base_structural_component import BaseStructuralComponent


class DockStream(BaseStructuralComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._configuration_path = self.parameters.specific_parameters[self.component_specific_parameters.DOCKSTREAM_CONFPATH]
        self._docker_script_path = self.parameters.specific_parameters[self.component_specific_parameters.DOCKSTREAM_DOCKERSCRIPTPATH]
        self._environment_path = self.parameters.specific_parameters[self.component_specific_parameters.DOCKSTREAM_ENVPATH]

    def _add_debug_mode_if_selected(self, command):
        if self.parameters.specific_parameters.get(self.component_specific_parameters.DOCKSTREAM_DEBUG, False)\
                or _is_development_environment():
            command = ' '.join([command, "-debug"])
        return command

    def _create_command(self, smiles: List[str], step):
        concat_smiles = '"' + ';'.join(smiles) + '"'
        command = ' '.join([self._environment_path,
                            self._docker_script_path,
                            "-conf", self._configuration_path,
                            "-output_prefix", self._get_step_string(step),
                            "-smiles", concat_smiles,
                            "-print_scores"])

        # check, if DockStream is to be executed in debug mode, which will cause its loggers to print out
        # much more detailed information
        command = self._add_debug_mode_if_selected(command)
        return command

    def _calculate_score(self, smiles: List[str], step) -> np.array:
        # create the external command
        command = self._create_command(smiles, step)

        # send the batch smiles and retrieve the result as a list of strings
        results = self._send_request_with_stepwize_read(command, len(smiles))

        # note: some ligands might have failed in DockStream (embedding or docking) although they are valid
        #       RDkit molecules -> "docker.py" will return "NA"'s for failed molecules, as '0' could be a perfectly
        #       normal value; anything that cannot be cast to a floating point number will result in '0'
        scores = []
        for score in results:
            try:
                score = float(score)
            except ValueError:
                score = 0
            scores.append(score)
        transformed_scores = self._transformation_function(scores, self.parameters.specific_parameters)

        return np.array(transformed_scores), np.array(scores)

    def _parse_result(self, result):
        return str(result).strip()



from typing import List
from dataclasses import dataclass


@dataclass
class ComponentParameters:
    component_type: str
    name: str
    weight: float
    model_path: str = None
    specific_parameters: dict = None



from reinvent_scoring.scoring.diversity_filters.lib_invent import NoFilter, NoFilterWithPenalty
from reinvent_scoring.scoring.diversity_filters.lib_invent.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.lib_invent.diversity_filter_parameters import DiversityFilterParameters


class DiversityFilter:

    def __new__(cls, parameters: DiversityFilterParameters) -> BaseDiversityFilter:
        all_filters = dict(NoFilter=NoFilter,
                           NoFilterWithPenalty=NoFilterWithPenalty)
        div_filter = all_filters.get(parameters.name, KeyError(f"Invalid filter name: `{parameters.name}'"))
        return div_filter(parameters)




class TransformationTypeEnum():
    __DOUBLE_SIGMOID = "double_sigmoid"
    __SIGMOID = "sigmoid"
    __REVERSE_SIGMOID = "reverse_sigmoid"
    __RIGHT_STEP = "right_step"
    __LEFT_STEP = "left_step"
    __STEP = "step"
    __CUSTOM_INTERPOLATION = "custom_interpolation"
    __NO_TRANSFORMATION = "no_transformation"

    @property
    def DOUBLE_SIGMOID(self):
        return self.__DOUBLE_SIGMOID

    @DOUBLE_SIGMOID.setter
    def DOUBLE_SIGMOID(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def SIGMOID(self):
        return self.__SIGMOID

    @SIGMOID.setter
    def SIGMOID(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def REVERSE_SIGMOID(self):
        return self.__REVERSE_SIGMOID

    @REVERSE_SIGMOID.setter
    def REVERSE_SIGMOID(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def RIGHT_STEP(self):
        return self.__RIGHT_STEP

    @RIGHT_STEP.setter
    def RIGHT_STEP(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def LEFT_STEP(self):
        return self.__LEFT_STEP

    @LEFT_STEP.setter
    def LEFT_STEP(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def STEP(self):
        return self.__STEP

    @STEP.setter
    def STEP(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def CUSTOM_INTERPOLATION(self):
        return self.__CUSTOM_INTERPOLATION

    @CUSTOM_INTERPOLATION.setter
    def CUSTOM_INTERPOLATION(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")

    @property
    def NO_TRANSFORMATION(self):
        return self.__NO_TRANSFORMATION

    @NO_TRANSFORMATION.setter
    def NO_TRANSFORMATION(self, value):
        raise ValueError("Do not assign value to a TransformationTypeEnum field")


class ContainerType:
    _SCIKIT_CONTAINER = "scikit_container"
    _OPTUNA_CONTAINER = "optuna_container"

    @property
    def SCIKIT_CONTAINER(self):
        return self._SCIKIT_CONTAINER

    @SCIKIT_CONTAINER.setter
    def SCIKIT_CONTAINER(self, value):
        raise ValueError("Do not assign value to a ContainerType field")

    @property
    def OPTUNA_CONTAINER(self):
        return self._OPTUNA_CONTAINER

    @OPTUNA_CONTAINER.setter
    def OPTUNA_CONTAINER(self, value):
        raise ValueError("Do not assign value to a ContainerType field")



from reinvent_scoring.scoring.diversity_filters.reinvent_core import IdenticalMurckoScaffold, \
    IdenticalTopologicalScaffold, ScaffoldSimilarity, NoScaffoldFilter
from reinvent_scoring.scoring.diversity_filters.reinvent_core.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.reinvent_core.diversity_filter_parameters import \
    DiversityFilterParameters


class DiversityFilter:

    def __new__(cls, parameters: DiversityFilterParameters) -> BaseDiversityFilter:
        all_filters = dict(IdenticalMurckoScaffold=IdenticalMurckoScaffold,
                           IdenticalTopologicalScaffold=IdenticalTopologicalScaffold,
                           ScaffoldSimilarity=ScaffoldSimilarity,
                           NoFilter=NoScaffoldFilter)
        div_filter = all_filters.get(parameters.name, KeyError(f"Invalid filter name: `{parameters.name}'"))
        return div_filter(parameters)



class ROCSInputFileTypesEnum():
    _SHAPE_QUERY = "shape_query"
    _SDF_QUERY = "sdf"
    # Need to stick to this pattern of getter/setter as the alternative more concise one cannot
    # be pickled for multiprocessing

    @property
    def SHAPE_QUERY(self):
        return self._SHAPE_QUERY

    @SHAPE_QUERY.setter
    def SHAPE_QUERY(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")

    @property
    def SDF_QUERY(self):
        return self._SDF_QUERY

    @SDF_QUERY.setter
    def SDF_QUERY(self, value):
        raise ValueError("Do not assign value to a ROCSInputFileTypesEnum field")




from abc import ABC, abstractmethod
from typing import List

from reinvent_chemistry.conversions import Conversions

from reinvent_scoring.scoring.enums import TransformationTypeEnum
from reinvent_scoring.scoring.score_transformations import TransformationFactory
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_summary import ComponentSummary
from reinvent_scoring.scoring.enums import ComponentSpecificParametersEnum


class BaseScoreComponent(ABC):

    def __init__(self, parameters: ComponentParameters):
        self.component_specific_parameters = ComponentSpecificParametersEnum()
        self.parameters = parameters
        self._chemistry = Conversions()
        self._transformation_function = self._assign_transformation(self.parameters.specific_parameters)

    @abstractmethod
    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        raise NotImplementedError("calculate_score method is not implemented")

    def calculate_score_for_step(self, molecules: List, step=-1) -> ComponentSummary:
        return self.calculate_score(molecules)

    def _assign_transformation(self, specific_parameters: {}):
        transformation_type = TransformationTypeEnum()
        factory = TransformationFactory()
        if not self.parameters.specific_parameters: #FIXME: this is a hack
            self.parameters.specific_parameters = {}
        if self.parameters.specific_parameters.get(self.component_specific_parameters.TRANSFORMATION, False):
            transform_function = factory.get_transformation_function(specific_parameters)
        else:
            self.parameters.specific_parameters[
                self.component_specific_parameters.TRANSFORMATION_TYPE] = transformation_type.NO_TRANSFORMATION
            transform_function = factory.no_transformation
        return transform_function


from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.pip.base_pip_model_component import BasePiPModelComponent


class PiPPredictionComponent(BasePiPModelComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def _parse_single_compound(self, compound):
        return float(compound["prediction"])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="reinvent_scoring",
    version="0.0.46",
    author="MolecularAI",
    author_email="patronov@gmail.com",
    description="Scoring functions for Reinvent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.astrazeneca.net/scm/com/reinvent_scoring.git",
    package_data={"reinvent_scoring": ["scoring/score_components/synthetic_accessibility/fpscores.pkl.gz"]},
    packages=setuptools.find_packages(exclude='unittest_reinvent'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



import io
import subprocess
from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class BaseConsoleInvokedComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def calculate_score_for_step(self, molecules: List, step=-1) -> ComponentSummary:
        return self.calculate_score(molecules, step)

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score, raw_score = self._calculate_score(valid_smiles, step)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)
        return score_summary

    def _get_step_string(self, step) -> str:
        if step == -1:
            return "\"\""
        return "".join(["\"e", str(step).zfill(4), "_\""])

    @abstractmethod
    def _calculate_score(self, smiles: List[str], step) -> np.array:
        raise NotImplementedError("_calculate_score method is not implemented")

    @abstractmethod
    def _create_command(self, step, input_json_path: str, output_json_path: str):
        raise NotImplementedError("_create_command method is not implemented")




from typing import Dict, Any

from reinvent_scoring.scoring.enums.container_type_enum import ContainerType
from reinvent_scoring.scoring.enums.component_specific_parameters_enum import ComponentSpecificParametersEnum

from reinvent_scoring.scoring.predictive_model.base_model_container import BaseModelContainer
from reinvent_scoring.scoring.predictive_model.optuna_container import OptunaModelContainer
from reinvent_scoring.scoring.predictive_model.scikit_model_container import ScikitModelContainer


class ModelContainer:

    def __new__(cls, activity_model: Any, specific_parameters: Dict) -> BaseModelContainer:
        _component_specific_parameters = ComponentSpecificParametersEnum()
        _container_type = ContainerType()
        container_type = specific_parameters.get(_component_specific_parameters.CONTAINER_TYPE,
                                                 _container_type.SCIKIT_CONTAINER)
        if container_type == _container_type.SCIKIT_CONTAINER:
            container_instance = ScikitModelContainer(activity_model,
                                                      specific_parameters[_component_specific_parameters.SCIKIT],
                                                      specific_parameters)
        else:
            # TODO: possibly a good spot for error try/catching
            container_instance = OptunaModelContainer(activity_model)

        return container_instance



import json
import os


def _is_development_environment() -> bool:
    try:
        project_root = os.path.dirname(__file__)
        with open(os.path.join(project_root, '../configs/config.json'), 'r') as f:
            config = json.load(f)
        is_dev = config.get("DEVELOPMENT_ENVIRONMENT", False)
        return is_dev
    except:
        return False



from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary


class BaseRESTComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self._request_url = self._create_url(self.parameters.component_type)
        self._request_header = self._create_header()

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score, raw_score = self._score_smiles(valid_smiles)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)

        return score_summary

    def _score_smiles(self, smiles: List[str]) -> np.array:
        response = self._post_request(self._request_url, smiles, self._request_header)
        results_raw = self._parse_response(response, len(smiles))
        results = self._apply_score_transformation(results_raw)

        return results, results_raw

    def _post_request(self, url, smiles, header):
        data = self._format_data(smiles)
        result = self._execute_request(url, data, header)

        return result

    @abstractmethod
    def _format_data(self, smiles: List[str]) -> dict:
        raise NotImplementedError("_format_data method is not implemented")

    @abstractmethod
    def _execute_request(self, request_url, data, header) -> dict:
        raise NotImplementedError("_execute_request method is not implemented")

    @abstractmethod
    def _parse_response(self, response_json: dict, data_size: int) -> np.array:
        raise NotImplementedError("_parse_response method is not implemented")

    def _apply_score_transformation(self, results_raw: np.array) -> np.array:
        """Returns np.array with non-NaN elements transformed by transformation function, and all NaN elements
        transformed into 0. """
        valid_mask = ~np.isnan(results_raw)
        results_raw_valid = results_raw[valid_mask]
        results_transformed = self._transformation_function(results_raw_valid, self.parameters.specific_parameters)
        results = np.zeros(len(results_raw), dtype=np.float32)
        results[valid_mask] = results_transformed

        return results

    @abstractmethod
    def _create_url(self, component_name) -> str:
        raise NotImplementedError("_create_url method is not implemented")

    @abstractmethod
    def _create_header(self) -> dict:
        raise NotImplementedError("_create_header method is not implemented")



from reinvent_scoring.scoring.enums.component_specific_parameters_enum import ComponentSpecificParametersEnum
from reinvent_scoring.scoring.enums.container_type_enum import ContainerType
from reinvent_scoring.scoring.enums.descriptor_types_enum import DescriptorTypesEnum
from reinvent_scoring.scoring.enums.environmental_variables_enum import EnvironmentalVariablesEnum
from reinvent_scoring.scoring.enums.logging_mode_enum import LoggingModeEnum
from reinvent_scoring.scoring.enums.rocs_input_file_types_enum import ROCSInputFileTypesEnum
from reinvent_scoring.scoring.enums.rocs_similarity_measures_enum import ROCSSimilarityMeasuresEnum
from reinvent_scoring.scoring.enums.rocs_specific_parameters_enum import ROCSSpecificParametersEnum
from reinvent_scoring.scoring.enums.scoring_function_component_enum import ScoringFunctionComponentNameEnum
from reinvent_scoring.scoring.enums.scoring_function_enum import ScoringFunctionNameEnum
from reinvent_scoring.scoring.enums.transformation_type_enum import TransformationTypeEnum



import abc
from typing import List

import numpy as np
import pandas as pd

from reinvent_scoring.scoring.diversity_filters.lib_invent.diversity_filter_memory import DiversityFilterMemory
from reinvent_scoring.scoring.diversity_filters.lib_invent.diversity_filter_parameters import DiversityFilterParameters
from reinvent_chemistry.conversions import Conversions


class BaseDiversityFilter(abc.ABC):

    @abc.abstractmethod
    def __init__(self, parameters: DiversityFilterParameters):
        self.parameters = parameters
        self._chemistry = Conversions()
        self._diversity_filter_memory = DiversityFilterMemory()

    @abc.abstractmethod
    def update_score(self, score_summary, sampled_sequences: List, step=0) -> np.array:
        raise NotImplementedError("The method 'evaluate' is not implemented!")

    def get_memory_as_dataframe(self) -> pd.DataFrame:
        return self._diversity_filter_memory.get_memory()

    def set_memory_from_dataframe(self, memory: pd.DataFrame):
        self._diversity_filter_memory.set_memory(memory)

    def number_of_smiles_in_memory(self) -> int:
        return self._diversity_filter_memory.number_of_smiles()

    def number_of_scaffold_in_memory(self) -> int:
        return self._diversity_filter_memory.number_of_scaffolds()

    def _calculate_scaffold(self, smile):
        raise NotImplementedError

    def _smiles_exists(self, smile):
        return self._diversity_filter_memory.smiles_exists(smile)

    def _add_to_memory(self, indx: int, score, smile, scaffold, components: List, step):
        self._diversity_filter_memory.update(indx, score, smile, scaffold, components, step)

    def _penalize_score(self, scaffold, score):
        """Penalizes the score if the scaffold bucket is full"""
        if self._diversity_filter_memory.scaffold_instances_count(scaffold) > self.parameters.bucket_size:
            score = 0.
        return score



import os
import multiprocessing
from multiprocessing import Pool
from pathlib import Path

import numpy as np
from openeye import oechem, oeomega, oeshape, oequacpac

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.enums import ROCSSimilarityMeasuresEnum, ROCSInputFileTypesEnum, ROCSSpecificParametersEnum
from reinvent_scoring.scoring.score_components.rocs import oehelper, oefuncs
from reinvent_scoring.scoring.score_components.rocs.base_rocs_component import BaseROCSComponent
from reinvent_scoring.scoring.score_components.rocs.default_values import ROCS_DEFAULT_VALUES


class ParallelRocsSimilarity(BaseROCSComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        avail_cpus = multiprocessing.cpu_count()
        oechem.OEThrow.SetLevel(10000)
        self.sim_measure_enum = ROCSSimilarityMeasuresEnum()
        self.input_types_enum = ROCSInputFileTypesEnum()
        self.param_names_enum = ROCSSpecificParametersEnum()

        self.num_cpus = min(avail_cpus, self._specific_param("MAX_CPUS"))
        self._set_omega_parameters()
        self._set_rocs_parameters()
        self.shape_weight = self._specific_param("SHAPE_WEIGHT")
        self.color_weight = self._specific_param("COLOR_WEIGHT")
        self.sim_func_name_set = oefuncs.get_similarity_name_set(parameters, self.param_names_enum,
                                                                 self.sim_measure_enum)

    def _set_omega_parameters(self):
        self.max_confs = self._specific_param("MAX_CONFS")
        self.erange = self._specific_param("EWINDOW")
        self.enum_stereo = self._specific_param("ENUM_STEREO")
        self.max_stereo = self._specific_param("MAX_STEREO")
        if self.max_stereo == 0:
            self.enum_stereo = False
        self.setup_omega(self.erange, self.max_confs)

    def _set_rocs_parameters(self):
        self.file_path = self._specific_param("ROCS_INPUT")
        self.file_type = self._specific_param("INPUT_TYPE")
        self.cff_path = self._specific_param("CUSTOM_CFF")
        self.save_overlays = self._specific_param("SAVE_ROCS_OVERLAYS")
        if self.save_overlays:
            self.dir_name = self._specific_param("ROCS_OVERLAYS_DIR")
            self.overlay_prefix = self._specific_param("ROCS_OVERLAYS_PREFIX")
            Path(self.dir_name).mkdir(parents=True, exist_ok=True)

        self.protein_file = ""
        self.ligand_file = ""
        self.neg_vol = self._specific_param("NEGATIVE_VOLUME")
        if self.neg_vol:
            self.protein_file = self._specific_param("PROTEIN_NEG_VOL_FILE")
            self.ligand_file = self._specific_param("LIGAND_NEG_VOL_FILE")

    def _calculate_omega_score(self, smiles, step) -> np.array:
        inputs = []
        if len(smiles) == 0:
            return np.array(())
        self._prepare_overlay()
        ind = str(step).zfill(4)

        for smile in smiles:
            input = {"smile": smile, "shape_weight": self.shape_weight, "color_weight": self.color_weight,
                     "sim_func_name_set": self.sim_func_name_set, "batch_id": ind,
                     "enum_stereo": self.enum_stereo, "max_stereo": self.max_stereo, "save_overlays": self.save_overlays,
                     "neg_vol_file": self.protein_file, "neg_vol_lig": self.ligand_file
                     }
            inputs.append(input)
        with Pool(processes=min(self.num_cpus, len(inputs))) as pool:
            results = pool.map(self._unfold, inputs)

        scores = []
        if self.save_overlays:
            overlay_filename = self.overlay_prefix + ind + ".sdf"
            overlay_file_path = os.path.join(self.dir_name, overlay_filename)
            outfs = oechem.oemolostream(overlay_file_path)
        for result in results:
            score, outmol = result
            scores.append(score)
            if self.save_overlays:
                oechem.OEWriteMolecule(outfs, outmol)
        return np.array(scores)

    def _prepare_overlay(self):
        overlay_function_types = {
            self.input_types_enum.SHAPE_QUERY: self.setup_reference_molecule_with_shape_query,
            self.input_types_enum.SDF_QUERY: self.setup_reference_molecule
        }
        overlay_function = overlay_function_types.get(self.file_type)
        overlay_function(self.file_path, self.cff_path)

    def _unfold(self, args):
        return self.parallel_scoring(**args)

    def _specific_param(self, key_enum):
        key = self.param_names_enum.__getattribute__(key_enum)
        default = ROCS_DEFAULT_VALUES[key_enum]
        ret = self.parameters.specific_parameters.get(key, default)
        if ret is not None:
            return ret
        raise KeyError(f"specific parameter \'{key}\' was not set")

    @classmethod
    def setup_reference_molecule_with_shape_query(cls, shape_query, cff_path):
        cls.prep = oeshape.OEOverlapPrep()
        qry = oeshape.OEShapeQuery()
        oefuncs.init_cff(cls.prep, cff_path)
        cls.rocs_overlay = oeshape.OEOverlay()
        if oeshape.OEReadShapeQuery(shape_query, qry):
            cls.rocs_overlay.SetupRef(qry)
        else:
            raise FileNotFoundError("A ROCS shape query file was not found")

    @classmethod
    def setup_reference_molecule(cls, file_path, cff_path):
        cls.prep = oeshape.OEOverlapPrep()
        input_stream = oechem.oemolistream()
        input_stream.SetFormat(oechem.OEFormat_SDF)
        input_stream.SetConfTest(oechem.OEAbsoluteConfTest(compTitles=False))
        refmol = oechem.OEMol()
        if input_stream.open(file_path):
            oechem.OEReadMolecule(input_stream, refmol)
        else:
            raise FileNotFoundError("A ROCS reference sdf file was not found")
        oefuncs.init_cff(cls.prep, cff_path)
        cls.prep.Prep(refmol)
        cls.rocs_overlay = oeshape.OEMultiRefOverlay()
        cls.rocs_overlay.SetupRef(refmol)

    @classmethod
    def setup_omega(cls, erange, max_confs):
        omegaOpts = oeomega.OEOmegaOptions()
        omegaOpts.SetStrictStereo(False)
        omegaOpts.SetEnergyWindow(erange)
        omegaOpts.SetMaxConfs(max_confs)
        cls.omega = oeomega.OEOmega(omegaOpts)
        return cls.omega

    @classmethod
    def parallel_scoring(cls, smile, shape_weight, color_weight, sim_func_name_set, batch_id, enum_stereo=False,
                         max_stereo=0, save_overlays=False, neg_vol_file="", neg_vol_lig=""):
        predicate = getattr(oeshape, sim_func_name_set.predicate)()
        imol = oechem.OEMol()
        outmol = oechem.OEMol()
        best_score = 0.0
        if oechem.OESmilesToMol(imol, smile):
            oequacpac.OEGetReasonableProtomer(imol)
            omega_success, imol = oehelper.get_omega_confs(imol, cls.omega, enum_stereo, max_stereo)
            if omega_success:
                cls.prep.Prep(imol)
                score = oeshape.OEBestOverlayScore()
                cls.rocs_overlay.BestOverlay(score, imol, predicate)
                outmol = oechem.OEGraphMol(imol.GetConf(oechem.OEHasConfIdx(score.GetFitConfIdx())))
                best_score, best_score_shape, best_score_color, neg_score = oehelper.get_score(outmol, score,
                                                                                               sim_func_name_set,
                                                                                               shape_weight,
                                                                                               color_weight,
                                                                                               neg_vol_file,
                                                                                               neg_vol_lig)
                if save_overlays:
                    oeshape.OERemoveColorAtoms(outmol)
                    oehelper.prep_sdf_file(outmol, score, smile, batch_id, best_score_shape, best_score_color,
                                           neg_score)
        return best_score, outmol



from reinvent_scoring.scoring.score_components.clab.azlogd74 import AZlogD74
from reinvent_scoring.scoring.score_components.clab.caco2_efflux import CACO2Efflux
from reinvent_scoring.scoring.score_components.clab.caco2_intr import CACO2Intrinsic
from reinvent_scoring.scoring.score_components.clab.herg import HERG
from reinvent_scoring.scoring.score_components.clab.hh_clint import HHClint
from reinvent_scoring.scoring.score_components.clab.hlm_clint import HLMClint
from reinvent_scoring.scoring.score_components.clab.rh_clint import RHClint
from reinvent_scoring.scoring.score_components.clab.solubilityDD import SolubilityDD
from reinvent_scoring.scoring.score_components.clab.top_20 import Top20





class ComponentSpecificParametersEnum:
    __LOW = "low"
    __HIGH = "high"
    __K = "k"
    __TRANSFORMATION = "transformation"
    __SCIKIT = "scikit"
    __CLAB_INPUT_FILE = "clab_input_file"
    __COEF_DIV = "coef_div"
    __COEF_SI = "coef_si"
    __COEF_SE = "coef_se"
    __TRANSFORMATION_TYPE = "transformation_type"
    __DESCRIPTOR_TYPE = "descriptor_type"
    __TRUNCATE_LEFT = "truncate_left"
    __TRUNCATE_RIGHT = "truncate_right"
    __INTERPOLATION_MAP = "interpolation_map"

    # structural components
    # ---------
    # AZDOCK
    __AZDOCK_CONFPATH = "configuration_path"
    __AZDOCK_DOCKERSCRIPTPATH = "docker_script_path"
    __AZDOCK_ENVPATH = "environment_path"
    __AZDOCK_DEBUG = "debug"

    # DockStream
    __DOCKSTREAM_CONFPATH = "configuration_path"
    __DOCKSTREAM_DOCKERSCRIPTPATH = "docker_script_path"
    __DOCKSTREAM_ENVPATH = "environment_path"
    __DOCKSTREAM_DEBUG = "debug"

    # AZGARD
    __AZGARD_CONFPATH = "configuration_path"
    __AZGARD_EXECUTORSCRIPTPATH = "executor_script_path"
    __AZGARD_ENVPATH = "environment_path"
    __AZGARD_VALUES_KEY = "values_key"
    __AZGARD_DEBUG = "debug"
    #######################

    __RAT_PK_PROPERTY = "rat_pk_property"
    __CLAB_TOP_20_VALUE = "clab_top_20_value"
    __ION_CLASS = "Ion class"
    __CONTAINER_TYPE = "container_type"

    __SMILES = "smiles"
    __MODEL_PATH = "model_path"

    @property
    def MODEL_PATH(self):
        return self.__MODEL_PATH

    @MODEL_PATH.setter
    def MODEL_PATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def SMILES(self):
        return self.__SMILES

    @SMILES.setter
    def SMILES(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def CONTAINER_TYPE(self):
        return self.__CONTAINER_TYPE

    @CONTAINER_TYPE.setter
    def CONTAINER_TYPE(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def ION_CLASS(self):
        return self.__ION_CLASS

    @ION_CLASS.setter
    def ION_CLASS(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def CLAB_TOP_20_VALUE(self):
        return self.__CLAB_TOP_20_VALUE

    @CLAB_TOP_20_VALUE.setter
    def CLAB_TOP_20_VALUE(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def RAT_PK_PROPERTY(self):
        return self.__RAT_PK_PROPERTY

    @RAT_PK_PROPERTY.setter
    def RAT_PK_PROPERTY(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZDOCK_DEBUG(self):
        return self.__AZDOCK_DEBUG

    @AZDOCK_DEBUG.setter
    def AZDOCK_DEBUG(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZDOCK_CONFPATH(self):
        return self.__AZDOCK_CONFPATH

    @AZDOCK_CONFPATH.setter
    def AZDOCK_CONFPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZDOCK_DOCKERSCRIPTPATH(self):
        return self.__AZDOCK_DOCKERSCRIPTPATH

    @AZDOCK_DOCKERSCRIPTPATH.setter
    def AZDOCK_DOCKERSCRIPTPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZDOCK_ENVPATH(self):
        return self.__AZDOCK_ENVPATH

    @AZDOCK_ENVPATH.setter
    def AZDOCK_ENVPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def DOCKSTREAM_DEBUG(self):
        return self.__DOCKSTREAM_DEBUG

    @DOCKSTREAM_DEBUG.setter
    def DOCKSTREAM_DEBUG(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def DOCKSTREAM_CONFPATH(self):
        return self.__DOCKSTREAM_CONFPATH

    @DOCKSTREAM_CONFPATH.setter
    def DOCKSTREAM_CONFPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def DOCKSTREAM_DOCKERSCRIPTPATH(self):
        return self.__DOCKSTREAM_DOCKERSCRIPTPATH

    @DOCKSTREAM_DOCKERSCRIPTPATH.setter
    def DOCKSTREAM_DOCKERSCRIPTPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def DOCKSTREAM_ENVPATH(self):
        return self.__DOCKSTREAM_ENVPATH

    @DOCKSTREAM_ENVPATH.setter
    def DOCKSTREAM_ENVPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZGARD_DEBUG(self):
        return self.__AZGARD_DEBUG

    @AZGARD_DEBUG.setter
    def AZGARD_DEBUG(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZGARD_CONFPATH(self):
        return self.__AZGARD_CONFPATH

    @AZGARD_CONFPATH.setter
    def AZGARD_CONFPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZGARD_EXECUTORSCRIPTPATH(self):
        return self.__AZGARD_EXECUTORSCRIPTPATH

    @AZGARD_EXECUTORSCRIPTPATH.setter
    def AZGARD_EXECUTORSCRIPTPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZGARD_VALUES_KEY(self):
        return self.__AZGARD_VALUES_KEY

    @AZGARD_VALUES_KEY.setter
    def AZGARD_VALUES_KEY(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def AZGARD_ENVPATH(self):
        return self.__AZGARD_ENVPATH

    @AZGARD_ENVPATH.setter
    def AZGARD_ENVPATH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def LOW(self):
        return self.__LOW

    @LOW.setter
    def LOW(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def HIGH(self):
        return self.__HIGH

    @HIGH.setter
    def HIGH(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def K(self):
        return self.__K

    @K.setter
    def K(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def TRANSFORMATION(self):
        return self.__TRANSFORMATION

    @TRANSFORMATION.setter
    def TRANSFORMATION(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def SCIKIT(self):
        return self.__SCIKIT

    @SCIKIT.setter
    def SCIKIT(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def CLAB_INPUT_FILE(self):
        return self.__CLAB_INPUT_FILE

    @CLAB_INPUT_FILE.setter
    def CLAB_INPUT_FILE(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def COEF_DIV(self):
        return self.__COEF_DIV

    @COEF_DIV.setter
    def COEF_DIV(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def COEF_SI(self):
        return self.__COEF_SI

    @COEF_SI.setter
    def COEF_SI(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def COEF_SE(self):
        return self.__COEF_SE

    @COEF_SE.setter
    def COEF_SE(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def TRANSFORMATION_TYPE(self):
        return self.__TRANSFORMATION_TYPE

    @TRANSFORMATION_TYPE.setter
    def TRANSFORMATION_TYPE(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def DESCRIPTOR_TYPE(self):
        return self.__DESCRIPTOR_TYPE

    @DESCRIPTOR_TYPE.setter
    def DESCRIPTOR_TYPE(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def TRUNCATE_LEFT(self):
        return self.__TRUNCATE_LEFT

    @TRUNCATE_LEFT.setter
    def TRUNCATE_LEFT(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def TRUNCATE_RIGHT(self):
        return self.__TRUNCATE_RIGHT

    @TRUNCATE_RIGHT.setter
    def TRUNCATE_RIGHT(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")

    @property
    def INTERPOLATION_MAP(self):
        return self.__INTERPOLATION_MAP

    @INTERPOLATION_MAP.setter
    def INTERPOLATION_MAP(self, value):
        raise ValueError("Do not assign value to a ComponentSpecificParametersEnum field")



from copy import deepcopy

import numpy as np
from rdkit import Chem
from rdkit.Chem.Scaffolds import MurckoScaffold
from reinvent_scoring.scoring.score_summary import FinalSummary

from reinvent_scoring.scoring.diversity_filters.reinvent_core.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.reinvent_core.diversity_filter_parameters import DiversityFilterParameters


class IdenticalMurckoScaffold(BaseDiversityFilter):
    """Penalizes compounds based on exact Murcko Scaffolds previously generated."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)

    def update_score(self, score_summary: FinalSummary, step=0) -> np.array:
        score_summary = deepcopy(score_summary)
        scores = score_summary.total_score
        smiles = score_summary.scored_smiles

        for i in score_summary.valid_idxs:
            smile = self._chemistry.convert_to_rdkit_smiles(smiles[i])
            scaffold = self._calculate_scaffold(smile)
            scores[i] = 0 if self._smiles_exists(smile) else scores[i]

            if scores[i] >= self.parameters.minscore:
                self._add_to_memory(i, scores[i], smile, scaffold, score_summary.scaffold_log, step)
                scores[i] = self._penalize_score(scaffold, scores[i])

        return scores

    def _calculate_scaffold(self, smile):
        mol = Chem.MolFromSmiles(smile)
        if mol:
            try:
                scaffold = MurckoScaffold.GetScaffoldForMol(mol)
                scaffold_smiles = Chem.MolToSmiles(scaffold, isomericSmiles=False)
            except ValueError:
                scaffold_smiles = ''
        else:
            scaffold_smiles = ''
        return scaffold_smiles



from copy import deepcopy

import numpy as np

from reinvent_scoring.scoring.diversity_filters.reinvent_core.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.reinvent_core.diversity_filter_parameters import DiversityFilterParameters
from reinvent_scoring.scoring.score_summary import FinalSummary


class NoScaffoldFilter(BaseDiversityFilter):
    """Don't penalize compounds."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)

    def update_score(self, score_summary: FinalSummary, step=0) -> np.array:
        score_summary = deepcopy(score_summary)
        scores = score_summary.total_score
        smiles = score_summary.scored_smiles
        for i in score_summary.valid_idxs:
            if scores[i] >= self.parameters.minscore:
                smile = self._chemistry.convert_to_rdkit_smiles(smiles[i])
                self._add_to_memory(i, scores[i], smile, smile, score_summary.scaffold_log, step)
        return scores



from openeye import oechem, oeshape, oeomega
from rdkit import Chem

def get_omega_confs(imol, omega, enum_stereo, max_stereo):
    stereo = False
    no_stereo = False
    if enum_stereo:
        enantiomers = list(oeomega.OEFlipper(imol.GetActive(), max_stereo, False, True))
        for k, enantiomer in enumerate(enantiomers):
            # Any other simpler way to combine and add all conformers to imol have failed !!
            # Failure = Creates conformers with wrong indices and wrong connections
            enantiomer = oechem.OEMol(enantiomer)
            ret_code = omega.Build(enantiomer)
            if ret_code == oeomega.OEOmegaReturnCode_Success:
                if k == 0:
                    imol = oechem.OEMol(enantiomer.SCMol())
                    imol.DeleteConfs()
                stereo = True
                for x in enantiomer.GetConfs():
                    imol.NewConf(x)
    else:
        no_stereo = omega(imol)
    return no_stereo or stereo, imol

def get_score(mol, score, sim_func_name_set, shape_weight, color_weight, neg_prot_file, neg_lig_file):
    neg_score = 0.0
    if len(neg_prot_file) > 0:
        neg_score = neg_vol_score(mol, neg_prot_file, neg_lig_file)

    best_score_shape = getattr(score, sim_func_name_set.shape)()
    best_score_shape = correct_shape_score(best_score_shape)
    best_score_shape = penalise_neg_volume(best_score_shape, neg_score)

    best_score_color = getattr(score, sim_func_name_set.color)()
    best_score_color = correct_color_score(best_score_color)
    best_score = ((shape_weight * best_score_shape) + (
            color_weight * best_score_color)) / (shape_weight + color_weight)
    return best_score, best_score_shape, best_score_color, neg_score

def neg_vol_score(mol, neg_prot_file, neg_lig_file):
    # 'mol' is the active conformation as obtained from overlay with the main query
    # It is important to use the protein ligand and do the overlay with mol again (query is the protein)
    # otherwise the score calculation is incorrect
    qfs = oechem.oemolistream()
    if not qfs.open(neg_lig_file):
        raise ValueError(f'Ligand file {neg_lig_file} could not be opened')
    qmol = oechem.OEMol()
    oechem.OEReadMolecule(qfs, qmol)

    efs = oechem.oemolistream()
    if not efs.open(neg_prot_file):
        raise ValueError(f'Protein file {neg_prot_file} could not be opened')
    emol = oechem.OEMol()
    oechem.OEReadMolecule(efs, emol)

    res = oeshape.OEROCSResult()
    evol = oeshape.OEExactShapeFunc()
    evol.SetupRef(emol)

    oeshape.OEROCSOverlay(res, qmol, mol)
    outmol = res.GetOverlayConf()

    # calculate overlap with protein
    eres = oeshape.OEOverlapResults()
    evol.Overlap(outmol, eres)

    frac = eres.GetOverlap() / eres.GetFitSelfOverlap()
    return frac

def prep_sdf_file(outmol, score, smile, batch_id, best_score_shape, best_score_color, neg_score):
    mol = Chem.MolFromSmiles(smile)
    smile = Chem.MolToSmiles(mol, canonical=True) if mol else ""
    score.Transform(outmol)
    oechem.OESetSDData(outmol, "Batch ID", batch_id)
    oechem.OESetSDData(outmol, "Smiles", smile)
    oechem.OESetSDData(outmol, "Shape", "%-.3f" % best_score_shape)
    oechem.OESetSDData(outmol, "Color", "%-.3f" % best_score_color)
    oechem.OESetSDData(outmol, "Negative vol penalty", "%-.3f" % neg_score)

def correct_color_score(score):
    if score >= 1.0:
        score = 0.90  # or alternative
    return score

def correct_shape_score(score):
    if score >= 1.0:
        score = 0.95  # or alternative
    return score

def penalise_neg_volume(score, neg_score):
    # Generally neg_score is not very high rouhly representing the % of molecule clashing with the protein
    # Assuming that even a low value is unacceptable, a different penalty function should be considered
    return score - neg_score


from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.clab.base_clab_component import BaseClabComponent


class Top20(BaseClabComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def _create_command(self, input_file) -> str:
        name = "clab_top20"
        command = f"ml {name}; {name} {input_file}"
        return command

    def _parse_compound(self, compound) -> float:
        top_20_value = self.parameters.specific_parameters[self.component_specific_parameters.CLAB_TOP_20_VALUE]
        if top_20_value == self.component_specific_parameters.ION_CLASS:
            ion_class = self.parameters.specific_parameters[top_20_value]
            result = any(compound[top_20_value] == c for c in ion_class)
            return float(result)
        return float(compound[top_20_value])



import pickle
from typing import List

import numpy as np
from rdkit.Chem import Mol
from rdkit.Chem.Descriptors import ExactMolWt

from reinvent_chemistry import Descriptors

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_components.synthetic_accessibility.sascorer import calculateScore
from reinvent_scoring.scoring.score_summary import ComponentSummary


class SASComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        self.activity_model = self._load_model(parameters)
        self._descriptors = Descriptors()
        self.fp_parameters = dict(
            radius=3,
            size=4096,  # Descriptors class calls this parameter "size", RDKit calls it "nBits".
            use_features=False,  # RDKit has False as default, Descriptors class has True.
        )

    def calculate_score(self, molecules: List[Mol]) -> ComponentSummary:
        score = self.predict_from_molecules(molecules)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters)
        return score_summary

    def predict_from_molecules(self, molecules: List[Mol]) -> np.ndarray:
        if len(molecules) == 0:
            return np.array([])

        descriptors = self._calculate_descriptors(molecules)

        # Normally, predict_proba takes a 2d array, one row per observation,
        # but a list of 1d arrays works too.
        sas_predictions = self.activity_model.predict_proba(descriptors)

        return sas_predictions[:, 1]

    def _load_model(self, parameters: ComponentParameters):
        try:
            model_path = self.parameters.specific_parameters.get(self.component_specific_parameters.MODEL_PATH, "")
            activity_model = self._load_scikit_model(model_path)
        except:
            model_path = self.parameters.specific_parameters.get(self.component_specific_parameters.MODEL_PATH, "")
            raise Exception(f"The loaded file `{model_path}` isn't a valid scikit-learn model")
        return activity_model

    def _load_scikit_model(self, model_path: str):
        with open(model_path, "rb") as f:
            scikit_model = pickle.load(f)
        return scikit_model

    def _calculate_descriptors(self, molecules: List[Mol]) -> List[np.ndarray]:
        descriptors = [self._sas_descriptor(mol) for mol in molecules]
        return descriptors

    def _sas_descriptor(self, mol: Mol) -> np.ndarray:
        """Returns SAS descriptor for a molecule, to be used as input to SAS model.

        SAS descriptor consists of three parts:
            1. SA score by Ertl and Schuffenhauer (Novartis), part of RDKit, copied to this repo.
            2. Molecular weight.
            3. Morgan fingerprint, with counts (ECFP6).

        The three parts are concatenated into one 1d numpy array.
        """

        sascore = calculateScore(mol)
        molwt = ExactMolWt(mol)
        fp = self._fingerprint(mol)

        descriptor = np.concatenate([[sascore], [molwt], fp])

        return descriptor

    def _fingerprint(self, mol: Mol) -> np.ndarray:
        fps = self._descriptors.molecules_to_count_fingerprints([mol], parameters=self.fp_parameters)
        return fps[0]
